/*
*
*	Example implementation for "Accurate Rendering of Liquid-Crystals and Inhomogeneous Optically Anisotropic Media"
*
* This code is licensed under the MIT license (MIT) (http://opensource.org/licenses/MIT)
*/


#version 450

#extension GL_ARB_separate_shader_objects : enable
#extension GL_ARB_shading_language_420pack : enable

layout (binding = 0) uniform restrict UBO {
	mat4 projection;
	mat4 modelView;
	mat4 invModelView;
	vec4 cameraPosition;
	vec4 lightPos;
	vec4 lightRadii;
	float wavelength;
	int white_light;
	int profile_type;
} ubo;
layout (binding = 1) uniform restrict Params {
	float t;
	float exposure;
	float tau;
	float polarization;
} params;

layout (location = 0) in vec3 color;
layout (location = 1) in vec3 position;
layout (location = 2) in vec3 wposition;

layout (location = 0) out vec4 outFragColor;

const float pi = 3.1415f;


float sqr(float x) { return x*x; }
// Adobe XYZ to RGB
vec3 XYZtoRGB(vec3 XYZ) {
	return XYZ * mat3( 2.0413690, -0.5649464, -0.3446944,
					  -0.9692660,  1.8760108,  0.0415560,
					   0.0134474, -0.1183897,  1.0154096);
}
vec3 white_point() { return vec3(0.9504, 1.0000, 1.0888); }

// Inputs:  Wavelength in nanometers
// From "Simple Analytic Approximations to the CIE XYZ Color Matching Functions"
float X_1931(float wave) {
	const float t1 = (wave-442.0f) * ((wave<442.0f)?0.0624f:0.0374f);
	const float t2 = (wave-599.8f) * ((wave<599.8f)?0.0264f:0.0323f);
	const float t3 = (wave-501.1f) * ((wave<501.1f)?0.0490f:0.0382f);
	return 0.362f * exp(-0.5f * sqr(t1)) + 1.056f * exp(-0.5f * sqr(t2)) - 0.065f * exp(-0.5f * sqr(t3));
}
float Y_1931(float wave) {
	const float t1 = (wave-568.8f) * ((wave<568.8f)?0.0213f:0.0247f);
	const float t2 = (wave-530.9f) * ((wave<530.9f)?0.0613f:0.0322f);
	return 0.821f * exp(-0.5f * sqr(t1)) + 0.286f * exp(-0.5f * sqr(t2));
}
float Z_1931(float wave) {
	const float t1 = (wave-437.0f) * ((wave<437.0f)?0.0845f:0.0278f);
	const float t2 = (wave-459.0f) * ((wave<459.0f)?0.0385f:0.0725f);
	return 1.217f * exp(-0.5f * sqr(t1)) + 0.681f * exp(-0.5f * sqr(t2));
}


vec3 materialcolor()
{
	return vec3(0.5,0.5,1);
}

// Normal Distribution function --------------------------------------
float D_GGX(float dotNH, float roughness)
{
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;
	float denom = dotNH * dotNH * (alpha2 - 1.0) + 1.0;
	return (alpha2)/(pi * denom*denom); 
}

// Geometric Shadowing function --------------------------------------
float G_SchlicksmithGGX(float dotNL, float dotNV, float roughness)
{
	float r = (roughness + 1.0);
	float k = (r*r) / 8.0;
	float GL = dotNL / (dotNL * (1.0 - k) + k);
	float GV = dotNV / (dotNV * (1.0 - k) + k);
	return GL * GV;
}

// Fresnel function ----------------------------------------------------
vec3 F_Schlick(float cosTheta, float metallic)
{
	vec3 F0 = mix(vec3(0.04), materialcolor(), metallic); // * material.specular
	vec3 F = F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0); 
	return F;    
}

// Specular BRDF composition --------------------------------------------

vec3 BRDF(vec3 L, vec3 V, vec3 N, float metallic, float roughness)
{
	// Precalculate vectors and dot products	
	vec3 H = normalize (V + L);
	float dotNV = clamp(dot(N, V), 0.0, 1.0);
	float dotNL = clamp(dot(N, L), 0.0, 1.0);
	float dotLH = clamp(dot(L, H), 0.0, 1.0);
	float dotNH = clamp(dot(N, H), 0.0, 1.0);

	// Light color fixed
	vec3 lightColor = vec3(1.0);

	vec3 color = vec3(0.0);

	if (dotNL > 0.0)
	{
		float rroughness = max(0.05, roughness);
		float D = D_GGX(dotNH, roughness); 
		float G = G_SchlicksmithGGX(dotNL, dotNV, roughness);
		vec3 F = F_Schlick(dotNV, metallic);

		vec3 spec = D * F * G / (4.0 * dotNL * dotNV);

		color += spec * dotNL * lightColor;
	}

	return color;
}

const vec3 lightPosition = vec3(0,-5,0);

void main() {
	const vec3 n = normalize(cross(dFdx(wposition),dFdy(wposition)));
	const vec3 o = normalize(-position * mat3(ubo.invModelView));
	const vec3 i = normalize(reflect(-o, n));
	
	const float wavelength = ubo.wavelength;
	const vec3 lightColor = XYZtoRGB(ubo.white_light>0 ? white_point() : vec3(X_1931(wavelength), Y_1931(wavelength), Z_1931(wavelength)));

	const float roughness = 0.05;
	const vec3 L = normalize(lightPosition - wposition);
	const vec3 Lo = BRDF(L, o, n, 0, roughness);

	vec3 color = materialcolor() * 0.005f;
	color += Lo * 0.25f;

	if (ubo.profile_type == 0) {
		const float d = length(wposition.xz - ubo.lightPos.xz);
		const vec2 R = abs(wposition.y - ubo.lightPos.y) * ubo.lightRadii.xy;
		const float cone = 1-clamp((d-R.x)/(R.y-R.x), .0f, 1.f);
		color += cone * lightColor.rgb * .15f;
	}

	outFragColor = vec4(color, .5f);
}

