/*
*
*	Example implementation for "Accurate Rendering of Liquid-Crystals and Inhomogeneous Optically Anisotropic Media"
*	 by Shlomi Steinberg
*	 http://ssteinberg.xyz
*
*
*    This program is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*
*/


#version 450

#extension GL_ARB_separate_shader_objects : enable
#extension GL_ARB_shading_language_420pack : enable

layout (constant_id = 0) const int view_mode = 0;
layout (constant_id = 1) const int layers = 0;
layout (constant_id = 2) const int white_light = 0;
layout (constant_id = 3) const int profile_type = 0;
layout (binding = 0) uniform restrict UBO {
	mat4 projection;
	mat4 modelView;
	mat4 invModelView;
	vec4 cameraPosition;
	vec4 lightPos;
	vec4 lightRadii;
	float wavelength;
} ubo;
layout (binding = 1) uniform restrict Params {
	float t;
	float exposure;
	float tau;
	float polarization;
	float c0, c1, c2, c3, c4;
} params;

layout (location = 0) in vec3 color;
layout (location = 1) in vec3 position;
layout (location = 2) in vec3 wposition;

layout (location = 0) out vec4 outFragColor;

const float pi = 3.1415926535897f;
const float eta_outside = 1.f;

const vec3 lightPosition = vec3(0,-5,0);


// Profile types
const int profile_slab = 0;
const int profile_droplet = 1;
const int profile_microdroplet = 2;
const int profile_droplet_axial = 3;
const int profile_droplet_pi_wall = 4;
const int profile_shell = 6;
const int profile_heat = 7;


// Lighting
vec3 materialcolor() { return vec3(.5f); }
float D_GGX(float dotNH, float roughness) {
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;
	float denom = dotNH * dotNH * (alpha2 - 1.0) + 1.0;
	return (alpha2)/(pi * denom*denom); 
}
float G_SchlicksmithGGX(float dotNL, float dotNV, float roughness) {
	float r = (roughness + 1.0);
	float k = (r*r) / 8.0;
	float GL = dotNL / (dotNL * (1.0 - k) + k);
	float GV = dotNV / (dotNV * (1.0 - k) + k);
	return GL * GV;
}
vec3 F_Schlick(float cosTheta, float metallic) {
	vec3 F0 = mix(vec3(0.04), materialcolor(), metallic); // * material.specular
	vec3 F = F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0); 
	return F;    
}
vec3 BRDF(vec3 L, vec3 V, vec3 N, float metallic, float roughness) {
	// Precalculate vectors and dot products	
	vec3 H = normalize (V + L);
	float dotNV = clamp(dot(N, V), 0.0, 1.0);
	float dotNL = clamp(dot(N, L), 0.0, 1.0);
	float dotLH = clamp(dot(L, H), 0.0, 1.0);
	float dotNH = clamp(dot(N, H), 0.0, 1.0);

	vec3 lightColor = vec3(1.0);
	vec3 color = vec3(0.0);
	if (dotNL > 0.0) {
		float rroughness = max(0.05, roughness);
		float D = D_GGX(dotNH, roughness); 
		float G = G_SchlicksmithGGX(dotNL, dotNV, roughness);
		vec3 F = F_Schlick(dotNV, metallic);

		vec3 spec = D * F * G / (4.0 * dotNL * dotNV);

		color += spec * dotNL * lightColor;
	}

	return color;
}


// Utility
float sqr(float x)			{ return x*x; }
double sqr(double x)		{ return x*x; }
vec2  sqr(vec2 x)			{ return x*x; }
vec3  sqr(vec3 x)			{ return x*x; }
vec4  sqr(vec4 x)			{ return x*x; }
vec2  cexpi(float phi) 		{ return vec2(cos(phi), sin(phi)); }
float max_element(vec2 v)	{ return max(v.x, v.y); }
float min_element(vec2 v)	{ return min(v.x, v.y); }
float max_element(vec3 v)	{ return max(v.x, max(v.y, v.z)); }
float min_element(vec3 v)	{ return min(v.x, min(v.y, v.z)); }
float max_element(vec4 v)	{ return max(max(v.x, v.y), max(v.w, v.z)); }
float min_element(vec4 v)	{ return min(min(v.x, v.y), min(v.w, v.z)); }
// Complex arithmetics
struct complex_mat2 { mat2 r, i; };
struct complex_vec2 { vec2 r, i; };
struct complex_dmat2 { dmat2 r, i; };
struct complex_dvec2 { dvec2 r, i; };
vec2  product(vec2 a, vec2 b) { return vec2(a.x*b.x-a.y*b.y, a.x*b.y+a.y*b.x); }
vec2  divide(vec2 a, vec2 b) { return vec2(a.x*b.x+a.y*b.y, a.y*b.x-a.x*b.y) / (b.x*b.x+b.y*b.y); }
dvec2 product(dvec2 a, dvec2 b) { return dvec2(a.x*b.x-a.y*b.y, a.x*b.y+a.y*b.x); }
dvec2 divide(dvec2 a, dvec2 b) { return dvec2(a.x*b.x+a.y*b.y, a.y*b.x-a.x*b.y) / (b.x*b.x+b.y*b.y); }
complex_vec2 cvflip(complex_vec2 v) {
	complex_vec2 ret;
	ret.r.xy = v.r.yx;
	ret.i.xy = v.i.yx;
	return ret;
}
complex_vec2 cvec2(vec2 a, vec2 b) {
	complex_vec2 v;
	v.r = vec2(a.x, b.x);
	v.i = vec2(a.y, b.y);
	return v;
}
complex_dvec2 cdvec2(dvec2 a, dvec2 b) {
	complex_dvec2 dv;
	dv.r = dvec2(a.x, b.x);
	dv.i = dvec2(a.y, b.y);
	return dv;
}
complex_vec2 cvec2(complex_dvec2 dv) {
	complex_vec2 v;
	v.r = vec2(dv.r);
	v.i = vec2(dv.i);
	return v;
}
complex_dvec2 cdvec2(complex_vec2 v) {
	complex_dvec2 dv;
	dv.r = vec2(v.r);
	dv.i = vec2(v.i);
	return dv;
}
complex_mat2 cidentity() {
	complex_mat2 R;
	R.r = mat2(1);
	R.i = mat2(0);
	return R;
}
complex_mat2 cmat2(vec2 a, vec2 b, vec2 c, vec2 d) {
	complex_mat2 R;
	R.r[0][0] = a.x;	R.i[0][0] = a.y;
	R.r[1][0] = b.x;	R.i[1][0] = b.y;
	R.r[0][1] = c.x;	R.i[0][1] = c.y;
	R.r[1][1] = d.x;	R.i[1][1] = d.y;
	return R;
}
complex_dmat2 cdmat2(dvec2 a, dvec2 b, dvec2 c, dvec2 d) {
	complex_dmat2 R;
	R.r[0][0] = a.x;	R.i[0][0] = a.y;
	R.r[1][0] = b.x;	R.i[1][0] = b.y;
	R.r[0][1] = c.x;	R.i[0][1] = c.y;
	R.r[1][1] = d.x;	R.i[1][1] = d.y;
	return R;
}
complex_mat2 cmat2(mat2 r, mat2 i) {
	complex_mat2 R;
	R.r = r; R.i = i;
	return R;
}
complex_dmat2 cdmat2(dmat2 r, dmat2 i) {
	complex_dmat2 R;
	R.r = r; R.i = i;
	return R;
}
complex_mat2 cmat2_add(complex_mat2 A, complex_mat2 B) {
	complex_mat2 R;
	R.r = A.r + B.r;
	R.i = A.i + B.i;
	return R;
}
complex_mat2 cmat2_substract(complex_mat2 A, complex_mat2 B) {
	complex_mat2 R;
	R.r = A.r - B.r;
	R.i = A.i - B.i;
	return R;
}
complex_mat2 cmat2_product(complex_mat2 A, complex_mat2 B) {
	const vec2 a = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d = vec2(A.r[1][1], A.i[1][1]);
	const vec2 a2 = vec2(B.r[0][0], B.i[0][0]);
	const vec2 b2 = vec2(B.r[1][0], B.i[1][0]);
	const vec2 c2 = vec2(B.r[0][1], B.i[0][1]);
	const vec2 d2 = vec2(B.r[1][1], B.i[1][1]);

	const vec2 Ra = product(a,a2) + product(b,c2);
	const vec2 Rb = product(a,b2) + product(b,d2);
	const vec2 Rc = product(c,a2) + product(d,c2);
	const vec2 Rd = product(c,b2) + product(d,d2);

	return cmat2(Ra, Rb, Rc, Rd);
}
complex_vec2 cmat2_by_cvec2(complex_mat2 A, complex_vec2 v) {
	const vec2 a  = vec2(v.r[0],v.i[0]);
	const vec2 b  = vec2(v.r[1],v.i[1]);
	const vec2 a2 = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b2 = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c2 = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d2 = vec2(A.r[1][1], A.i[1][1]);

	const vec2 Ra = product(a,a2) + product(b,b2);
	const vec2 Rb = product(a,c2) + product(b,d2);

	return cvec2(Ra, Rb);
}
complex_dvec2 cdmat2_by_cdvec2(complex_dmat2 A, complex_dvec2 v) {
	const dvec2 a  = dvec2(v.r[0],v.i[0]);
	const dvec2 b  = dvec2(v.r[1],v.i[1]);
	const dvec2 a2 = dvec2(A.r[0][0], A.i[0][0]);
	const dvec2 b2 = dvec2(A.r[1][0], A.i[1][0]);
	const dvec2 c2 = dvec2(A.r[0][1], A.i[0][1]);
	const dvec2 d2 = dvec2(A.r[1][1], A.i[1][1]);

	const dvec2 Ra = product(a,a2) + product(b,b2);
	const dvec2 Rb = product(a,c2) + product(b,d2);

	return cdvec2(Ra, Rb);
}
complex_vec2 mat2_by_cvec2(mat2 A, complex_vec2 v) {
	const vec2 a  = vec2(v.r[0],v.i[0]);
	const vec2 b  = vec2(v.r[1],v.i[1]);
	const vec2 a2 = vec2(A[0][0], 0);
	const vec2 b2 = vec2(A[1][0], 0);
	const vec2 c2 = vec2(A[0][1], 0);
	const vec2 d2 = vec2(A[1][1], 0);

	const vec2 Ra = product(a,a2) + product(b,b2);
	const vec2 Rb = product(a,c2) + product(b,d2);

	return cvec2(Ra, Rb);
}
complex_dvec2 dmat2_by_cdvec2(dmat2 A, complex_dvec2 v) {
	const dvec2 a  = dvec2(v.r[0],v.i[0]);
	const dvec2 b  = dvec2(v.r[1],v.i[1]);
	const dvec2 a2 = dvec2(A[0][0], 0);
	const dvec2 b2 = dvec2(A[1][0], 0);
	const dvec2 c2 = dvec2(A[0][1], 0);
	const dvec2 d2 = dvec2(A[1][1], 0);

	const dvec2 Ra = product(a,a2) + product(b,b2);
	const dvec2 Rb = product(a,c2) + product(b,d2);

	return cdvec2(Ra, Rb);
}
complex_mat2 cinv(complex_mat2 A) {
	const vec2 a = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d = vec2(A.r[1][1], A.i[1][1]);

	const vec2 N = divide(vec2(1,0), product(a,d) - product(b,c));
	const vec2 Ra = divide(d,N);
	const vec2 Rb = divide(-b,N);
	const vec2 Rc = divide(-c,N);
	const vec2 Rd = divide(a,N);

	return cmat2(Ra, Rb, Rc, Rd);
}
vec2  mag2(complex_vec2 v) 	{ return vec2(sqr(v.r[0]) + sqr(v.i[0]), sqr(v.r[1]) + sqr(v.i[1])); }


// Double-precision complex utilities
dvec2  conj(dvec2 a) 			{ return dvec2(a.x, -a.y); }
dvec2  cdprod(dvec2 a, dvec2 b) { return dvec2(a.x*b.x-a.y*b.y, a.x*b.y+a.y*b.x); }
dvec2  cdsqr(dvec2 a) 			{ return dvec2(a.x*a.x-a.y*a.y, 2*a.x*a.y); }
dvec2  cddiv(dvec2 a, dvec2 b) 	{ return dvec2(a.x*b.x+a.y*b.y, a.y*b.x-a.x*b.y) * (1.0lf/(b.x*b.x+b.y*b.y)); }
dvec2  cddiv(double a, dvec2 b) { return a*conj(b) * (1.0lf/(b.x*b.x+b.y*b.y)); }
double real(dvec2 z) 			{ return z.x; }
double imag(dvec2 z) 			{ return z.y; }
double cabs(dvec2 z) 			{ return sqrt(z.x*z.x + z.y*z.y); }
double argd(dvec2 z)			{ return double(atan(float(z.y), float(z.x))); }
dvec2  cexpdi(double d) 		{ return dvec2(cos(float(d)), sin(float(d))); }
double expd(double d) 			{ return exp(float(d)); }
dvec2  cexpd(dvec2 z) 			{ return expd(z.x) * cexpdi(z.y); }
double sqrtd(double z) 			{ return sqrt(max(.0lf,z)); }
double powd(double d, double a) { 
	const double sq = d>0?sqrt(d):0;
	const double d2 = d*d;
	const double d3 = d2*d;
	const double d4 = d2*d2;
	const double d5 = d3*d2;
	if (a == 0.0lf) return 1.0lf;
	if (a == 1.0lf) return d;
	if (a == 2.0lf) return d2;
	if (a == 3.0lf) return d3;
	if (a == 4.0lf) return d2*d2;
	if (a == 5.0lf) return d3*d2;
	if (a == 6.0lf) return d3*d3;
	if (a == 7.0lf) return d4*d3;
	if (a == 8.0lf) return d4*d4;
	if (a == 9.0lf) return d5*d4;
	if (a == 10.0lf) return d5*d5;
	if (a == 0.5lf) return sq;
	if (a == 1.5lf) return sq*d;
	if (a == 2.5lf) return sq*d2;
	if (a == 3.5lf) return sq*d3;
	if (a == 4.5lf) return sq*d2*d2;
	if (a == 5.5lf) return sq*d3*d2;
	if (a == 6.5lf) return sq*d3*d3;
	if (a == 7.5lf) return sq*d4*d3;
	if (a == 8.5lf) return sq*d4*d4;
	if (a == 9.5lf) return sq*d5*d4;
	if (a == 10.5lf) return sq*d5*d5;

	return pow(float(d),float(a));
}
dvec2  powd(dvec2 z, double a) {
	const dvec2 z2 = cdprod(z,z);
	const dvec2 z3 = cdprod(z2,z);
	const dvec2 z4 = cdprod(z2,z2);
	const dvec2 z5 = cdprod(z3,z2);
	if (a == 0.0lf) return dvec2(1,0);
	if (a == 1.0lf) return z;
	if (a == 2.0lf) return z2;
	if (a == 3.0lf) return z3;
	if (a == 4.0lf) return cdprod(z2,z2);
	if (a == 5.0lf) return cdprod(z3,z2);
	if (a == 6.0lf) return cdprod(z3,z3);
	if (a == 7.0lf) return cdprod(z4,z3);
	if (a == 8.0lf) return cdprod(z4,z4);
	if (a == 9.0lf) return cdprod(z5,z4);
	if (a == 10.0lf) return cdprod(z5,z5);

	const double r = cabs(z);
	const double phi = argd(z);
	return powd(r,a) * cexpdi(phi*a);
}
double expd2(double x) 			{ return powd(2.lf, x); }
dvec2 sqrtd(dvec2 z) 			{ return powd(z, 0.5lf); }
dvec2 csqrtd(dvec2 z) 			{ return powd(z, 0.5lf); }
dvec2 csqrtd(double d) {
	const double sd = sqrt(abs(d));
	return d > 0 ? dvec2(sd, 0) : dvec2(0, sd);
}
double cos(double r) { return double(cos(float(r))); }
double sin(double r) { return double(sin(float(r))); }

// The Faddeeva function
dvec2 faddeeva(dvec2 z) {
	const int M = 4;
	const int N = 1 << (M-1);

	// Precomputed Faddeeva function approximation coefficients (M==4)
	const dvec2 A[N] = {
		dvec2(+0.983046454995208lf, 0.0lf),
		dvec2(-0.095450491368505lf, 0.0lf),
		dvec2(-0.106397537035019lf, 0.0lf),
		dvec2(+0.004553979597404lf, 0.0lf),
		dvec2(-0.000012773721299lf, 0.0lf),
		dvec2(-0.000000071458742lf, 0.0lf),
		dvec2(+0.000000000080803lf, 0.0lf),
		dvec2(-0.000000000000007lf, 0.0lf),
	};
	const dvec2 B[N] = {
		dvec2(0.0lf, -1.338045597353875lf),
		dvec2(0.0lf, +0.822618936152688lf),
		dvec2(0.0lf, -0.044470795125534lf),
		dvec2(0.0lf, -0.000502542048995lf),
		dvec2(0.0lf, +0.000011914499129lf),
		dvec2(0.0lf, -0.000000020157171lf),
		dvec2(0.0lf, -0.000000000001558lf),
		dvec2(0.0lf, +0.000000000000003lf),
	};
	const dvec2 C[N] = {
		dvec2(0.392699081698724lf, 0.0lf),
		dvec2(1.178097245096172lf, 0.0lf),
		dvec2(1.963495408493621lf, 0.0lf),
		dvec2(2.748893571891069lf, 0.0lf),
		dvec2(3.534291735288517lf, 0.0lf),
		dvec2(4.319689898685965lf, 0.0lf),
		dvec2(5.105088062083414lf, 0.0lf),
		dvec2(5.890486225480862lf, 0.0lf),
	};
	const double s = 2.75lf;

	// Constrain to imag(z)>=0
	const double sgni = imag(z)<0 ? -1 : 1;
	z *= sgni;

	// Approximate
	const dvec2 t = z + dvec2(0,0.5lf)*s;
	dvec2 w = dvec2(.0);
	for (int m=0; m<N; ++m)
		w += cddiv(A[m] + cdprod(t,B[m]), cdsqr(C[m]) - cdsqr(t));

	// Invert back
	if (sgni < 0)
		w = 2.0lf*cexpd(-cdsqr(z)) - w;

	return w;
}
// Error function
dvec2 erf(dvec2 z) {
	dvec2 z_1i = cdprod(dvec2(0,1), z);	// 1i*z
	return dvec2(1,0) - cdprod(cexpd(-cdsqr(z)), faddeeva(z_1i));
}
dvec2 Ig_erf0(dvec2 a, dvec2 b, double y) {
    return 0.5lf * (erf((a+b*y)/sqrt(2.lf)) - erf(a/sqrt(2.lf)));
}


// Adobe XYZ to RGB
// (used as an example)
vec3 XYZtoRGB(vec3 XYZ) {
	return XYZ * mat3(  2.9515373, -1.2894116, -0.4738445,
					   -1.0851093,  1.9908566,  0.0372026,
						0.0854934, -0.2694964,  1.0912975);
}

// Inputs:  Wavelength in nanometers
// From "Simple Analytic Approximations to the CIE XYZ Color Matching Functions"
float X_1931(float wave) {
	const float t1 = (wave-442.0f) * ((wave<442.0f)?0.0624f:0.0374f);
	const float t2 = (wave-599.8f) * ((wave<599.8f)?0.0264f:0.0323f);
	const float t3 = (wave-501.1f) * ((wave<501.1f)?0.0490f:0.0382f);
	return 0.362f * exp(-0.5f * sqr(t1)) + 1.056f * exp(-0.5f * sqr(t2)) - 0.065f * exp(-0.5f * sqr(t3));
}
float Y_1931(float wave) {
	const float t1 = (wave-568.8f) * ((wave<568.8f)?0.0213f:0.0247f);
	const float t2 = (wave-530.9f) * ((wave<530.9f)?0.0613f:0.0322f);
	return 0.821f * exp(-0.5f * sqr(t1)) + 0.286f * exp(-0.5f * sqr(t2));
}
float Z_1931(float wave) {
	const float t1 = (wave-437.0f) * ((wave<437.0f)?0.0845f:0.0278f);
	const float t2 = (wave-459.0f) * ((wave<459.0f)?0.0385f:0.0725f);
	return 1.217f * exp(-0.5f * sqr(t1)) + 0.681f * exp(-0.5f * sqr(t2));
}
vec3 XYZ_1931(float wavelength) { return vec3(X_1931(wavelength), Y_1931(wavelength), Z_1931(wavelength)); }
vec3 white_point() { return vec3(0.9504, 1.0000, 1.0888); }


// Normal modes
float qo(float eo, float K) {
	return sqrt(sqr(eo) - sqr(K));
}
double qo(double eo, double K) {
	return sqrt(sqr(eo) - sqr(K));
}
vec2 qe(float eo, float ee, vec3 A, float K) {
    const float e = sqr(ee)-sqr(eo);
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float G = sqr(b)*e+sqr(eo);

	return vec2(c*b*K*e + eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(ee)*G),
				c*b*K*e - eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(ee)*G))/G;
}
dvec2 qe(double eo, double ee, dvec3 A, double K) {
    const double e = sqr(ee)-sqr(eo);
    const double a = A.x;
    const double b = A.y;
    const double c = A.z;
    const double G = sqr(b)*e+sqr(eo);

	return dvec2(c*b*K*e + eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(ee)*G),
				 c*b*K*e - eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(ee)*G))/G;
}
// Electric fields
vec3 E_e(float eo, float q, vec3 A, float K) {
	const vec3 We = vec3(0,q,-K);
	const float AWe = dot(A, We);
	return (sqr(eo)*A - AWe*We) / sqrt(sqr(sqr(eo)) + sqr(AWe)*(sqr(q) + sqr(K) - 2*sqr(eo)));
}
dvec3 E_e(double eo, double q, dvec3 A, double K) {
	const dvec3 We = dvec3(0,q,-K);
	const double AWe = dot(A, We);
	return (sqr(eo)*A - AWe*We) / sqrt(sqr(sqr(eo)) + sqr(AWe)*(sqr(q) + sqr(K) - 2*sqr(eo)));
}
vec3 E_ep(float eo, float ee, vec3 A, float K) { return E_e(eo, qe(eo, ee, A, K).x, A, K); }
dvec3 E_ep(double eo, double ee, dvec3 A, double K) { return E_e(eo, qe(eo, ee, A, K).x, A, K); }
vec3 E_em(float eo, float ee, vec3 A, float K) { return E_e(eo, qe(eo, ee, A, K).y, A, K); }
dvec3 E_em(double eo, double ee, dvec3 A, double K) { return E_e(eo, qe(eo, ee, A, K).y, A, K); }
vec3 E_o(float eo, float q, vec3 A, float K) {
	const vec3 Wo = vec3(0,q,-K);
	return cross(A,Wo) / sqrt(sqr(eo) - sqr(dot(A,Wo)));
}
dvec3 E_o(double eo, double q, dvec3 A, double K) {
	const dvec3 Wo = dvec3(0,q,-K);
	return cross(A,Wo) / sqrt(sqr(eo) - sqr(dot(A,Wo)));
}
vec3 E_op(float eo, vec3 A, float K) { return E_o(eo,  qo(eo, K), A, K); }
dvec3 E_op(double eo, dvec3 A, double K) { return E_o(eo,  qo(eo, K), A, K); }
vec3 E_om(float eo, vec3 A, float K) { return E_o(eo, -qo(eo, K), A, K); }
dvec3 E_om(double eo, dvec3 A, double K) { return E_o(eo, -qo(eo, K), A, K); }

// Fresnel coefficients
void fresnel_iso_aniso(float cos_theta, float sin_theta, vec3 A, float ei, float eo, float ee,
					   out float rss, out float rsp, out float tso, out float tse, out float rps, out float rpp, out float tpo, out float tpe) {
	const float K = ei*sin_theta;
	const float qi = -qo(ei,K);
	const vec3 E2om = E_om(eo,A,K);
	const vec3 E2em = E_em(eo,ee,A,K);
	const float q2om = -qo(eo,K);
	const float q2em = qe(eo,ee,A,K).y;

	const float A_1 = -E2em.x*E2om.y*K*cos_theta;
	const float A_2 = -E2em.y*E2om.x*K*cos_theta;
	const float C_1 = qi*cos_theta - K*sin_theta;
	const float C_2 = E2em.x*E2om.z*(C_1 + q2om*cos_theta) - A_1;
	const float C_3 = E2em.z*E2om.x*(C_1 + q2em*cos_theta) - A_2;

	const float Ns = C_2*(qi + q2em) - C_3*(qi + q2om);

	rps = (2*E2em.x*E2om.x*cos_theta*C_1*(q2om - q2em)) / Ns;
	rpp = (-E2em.x*E2om.z*(C_1 - q2om*cos_theta)*(qi + q2em) + E2em.z*E2om.x*(C_1 - q2em*cos_theta)*(qi + q2om) - A_1*(q2em + qi) + A_2*(q2om + qi)) / Ns;
	tpo = -(2*E2em.x*cos_theta*(q2em + qi)*C_1) / Ns;
	tpe = (2*E2om.x*cos_theta*(q2om + qi)*C_1) / Ns;

	rss = (C_2*(qi - q2em) + C_3*(q2om - qi)) / Ns;
	rsp = (2*qi*(E2em.z*E2om.y*K - E2em.y*E2om.z*K + E2em.z*E2om.z*(q2om - q2em))) / Ns;
	tso = -(2*qi*(E2em.y*K*cos_theta + E2em.z*C_1 + E2em.z*q2em*cos_theta)) / Ns;
	tse = (2*qi*(E2om.y*K*cos_theta + E2om.z*C_1 + E2om.z*q2om*cos_theta)) / Ns;
}
void fresnel_aniso_iso(float cos_phi, float sin_phi, vec3 A, float eo, float ee, float eo2,
					   out float roo, out float roe, out float tos, out float top, out float reo, out float ree, out float tes, out float tep) {
	const float K = eo2*sin_phi;
	const float q2om = -qo(eo2,K);
	const float qop  =  qo(eo,K);
	const vec2 _qe = qe(eo,ee,A,K);
	const float qep = _qe.x;
	const float qem = _qe.y;
	const vec3 Eop = E_op(eo,A,K);
	const vec3 Eep = E_ep(eo,ee,A,K);
	const vec3 Eom = E_om(eo,A,K);
	const vec3 Eem = E_em(eo,ee,A,K);

	const float B_1 = -Eep.x*K*cos_phi*(q2om - qep);
	const float B_2 = -Eom.x*K*cos_phi*(q2om + qop);
	const float B_3 = -Eop.x*K*cos_phi*(qop - q2om);
	const float D_1 = K*sin_phi - q2om*cos_phi;
	const float D_2 = Eep.z*(D_1 + qep*cos_phi);
	const float D_3 = Eem.z*(D_1 + qem*cos_phi);
	const float D_4 = Eop.z*(D_1 + qop*cos_phi);

	const float N2 = (Eep.x*(qep - q2om)*D_4 + Eop.x*(q2om - qop)*D_2 + Eop.y*B_1 + Eep.y*B_3);

	roo = -(Eep.x*Eom.z*(q2om - qep)*(qop*cos_phi - D_1) + Eom.x*(q2om + qop)*D_2 + Eom.y*B_1 - Eep.y*B_2) / N2;
	roe = -(-Eom.x*Eop.z*(qop*(qop*cos_phi + K*sin_phi) + q2om*D_1) + Eom.z*Eop.x*(qop*(qop*cos_phi - K*sin_phi) + q2om*D_1) + Eop.y*B_2 + Eom.y*B_3) / N2;
	tos = (Eep.x*qep*roe - Eom.x*qop + Eop.x*qop*roo) / q2om;
	top = -(Eom.z*qop - Eom.y*K - (Eep.y*K + Eep.z*qep)*roe - (Eop.y*K + Eop.z*qop)*roo) / D_1;

	reo = (Eem.x*(qem - q2om)*(D_2 + Eep.y*K*cos_phi) + Eep.x*(q2om - qep)*(D_3 + Eem.y*K*cos_phi)) / N2;
	ree = -(Eem.x*(qem - q2om)*D_4 - Eem.x*Eop.y*K*cos_phi*(q2om - qem) + Eem.y*B_3 + Eop.x*(q2om - qop)*D_3) / N2;
	tes = (Eem.x*qem + Eep.x*qep*ree + Eop.x*qop*reo) / q2om;
	tep = -(Eem.z + Eep.z*ree + Eop.z*reo) / cos_phi;
}
void fresnel_aniso_aniso(float K, float qi, vec3 Ei, float eo, float eo2, float ee, float ee2, vec3 A, vec3 A2,
						 out float ro, out float re, out float to, out float te) {
    const vec3 Eop = E_op(eo,A,K);
    const vec3 Eep = E_ep(eo,ee,A,K);
    const float qop = qo(eo,K);
    const float qep = qe(eo,ee,A,K).x;
    const vec3 E2om = E_om(eo2,A2,K);
    const vec3 E2em = E_em(eo2,ee2,A2,K);
    const float q2om = -qo(eo2,K);
    const float q2em = qe(eo2,ee2,A2,K).y;
    
    const float C_1 = q2em - q2om;
    const float C_2 = qep - qi;
    const float C_3 = q2em - qi;
    const float C_4 = q2em - qep;
    const float C_5 = q2om - qi;
    const float C_6 = q2om - qep;
    const float C_7 = qep - qop;
    const float C_8 = q2em - qop;
    const float C_9 = q2om - qop;

	const float Nro = 
		  E2em.x*E2om.x*C_1*(Eep.y*Eop.z*K - Eep.z*Eop.y*K + Eep.z*Eop.z*C_7)
		+ E2em.x*E2om.y*K*(Eep.z*Eop.x*C_8 - Eep.x*Eop.z*C_4)
		+ E2em.x*E2om.z*(Eop.x*C_8*(Eep.z*C_6 - Eep.y*K) + Eep.x*C_4*(Eop.y*K - Eop.z*C_9))
		+ E2em.y*E2om.x*K*(Eep.x*Eop.z*C_6 - Eep.z*Eop.x*C_9) 
		+ Eep.x*Eop.x*C_7*(K*E2em.y*E2om.z - K*E2em.z*E2om.y + E2em.z*E2om.z*C_1)
		+ E2em.z*E2om.x*(Eep.x*C_6*(Eop.z*C_8 - Eop.y*K) + Eop.x*C_9*(Eep.y*K - Eep.z*C_4));
    ro = 
		(-E2em.x*E2om.y*K*(Eep.z*Ei.x*C_3 - Eep.x*Ei.z*C_4)
		+ E2em.x*E2om.z*(Eep.x*C_4*(Ei.z*C_5 - Ei.y*K) - Ei.x*C_3*(Eep.z*C_6 - Eep.y*K))
		- E2em.y*E2om.x*K*(Eep.x*Ei.z*C_6 - Eep.z*Ei.x*C_5)
		- Eep.x*Ei.x*K*C_2*(E2em.y*E2om.z - E2em.z*E2om.y)
		+ E2em.z*E2om.x*(Ei.x*C_5*(Eep.z*C_4 - Eep.y*K) - Eep.x*C_6*(Ei.z*C_3 - Ei.y*K))
		- E2em.z*E2om.z*(Eep.x*Ei.x*q2em*C_2 - Eep.x*Ei.x*q2om*C_2)
		+ E2em.x*E2om.x*C_1*(Eep.z*Ei.y*K - Eep.y*Ei.z*K - Eep.z*Ei.z*C_2))
		/ Nro;

	const float Nre = E2om.z*(E2em.x*q2em*(Eep.z*C_6 - Eep.y*K) + Eep.x*qep*(E2em.z*C_1 + E2em.y*K))
		- E2om.x*q2om*(E2em.z*(Eep.z*C_4 - Eep.y*K) + E2em.y*Eep.z*K)
		- E2om.y*K*(E2em.z*Eep.x*qep - E2em.x*Eep.z*q2em);
    re = abs(Nre) < 1e-8f ? .0f : (
		-(E2em.x*E2om.z*q2em*(Ei.z*C_5 - Ei.y*K - (Eop.y*K - Eop.z*C_9)*ro) 
		- E2em.z*E2om.x*q2om*(Ei.z*C_3 - Ei.y*K - (Eop.y*K - Eop.z*C_8)*ro) 
		+ E2em.z*E2om.z*C_1*(Ei.x*qi + Eop.x*qop*ro) 
		- K*(Ei.z + Eop.z*ro)*(E2em.y*E2om.x*q2om - E2em.x*E2om.y*q2em)
		- K*(Ei.x*qi + Eop.x*qop*ro)*(E2em.z*E2om.y - E2em.y*E2om.z))
		/ Nre);

    const float Nt = E2em.x*E2om.z - E2em.z*E2om.x;
    const float D_1 = Ei.z + Eep.z*re + Eop.z*ro;
    const float D_2 = Ei.x + Eep.x*re + Eop.x*ro;

    to = (E2em.x*D_1 - E2em.z*D_2) / Nt;
    te = (E2om.z*D_2 - E2om.x*D_1) / Nt;
}
void fresnel_aniso_aniso(double K, double qi, dvec3 Ei, double eo, double eo2, double ee, double ee2, dvec3 A, dvec3 A2,
						 out double ro, out double re, out double to, out double te) {
    const dvec3 Eop = E_op(eo,A,K);
    const dvec3 Eep = E_ep(eo,ee,A,K);
    const double qop = qo(eo,K);
    const double qep = qe(eo,ee,A,K).x;
    const dvec3 E2om = E_om(eo2,A2,K);
    const dvec3 E2em = E_em(eo2,ee2,A2,K);
    const double q2om = -qo(eo2,K);
    const double q2em = qe(eo2,ee2,A2,K).y;
    
    const double C_1 = q2em - q2om;
    const double C_2 = qep - qi;
    const double C_3 = q2em - qi;
    const double C_4 = q2em - qep;
    const double C_5 = q2om - qi;
    const double C_6 = q2om - qep;
    const double C_7 = qep - qop;
    const double C_8 = q2em - qop;
    const double C_9 = q2om - qop;

	const double Nro = 
		  E2em.x*E2om.x*C_1*(Eep.y*Eop.z*K - Eep.z*Eop.y*K + Eep.z*Eop.z*C_7)
		+ E2em.x*E2om.y*K*(Eep.z*Eop.x*C_8 - Eep.x*Eop.z*C_4)
		+ E2em.x*E2om.z*(Eop.x*C_8*(Eep.z*C_6 - Eep.y*K) + Eep.x*C_4*(Eop.y*K - Eop.z*C_9))
		+ E2em.y*E2om.x*K*(Eep.x*Eop.z*C_6 - Eep.z*Eop.x*C_9) 
		+ Eep.x*Eop.x*C_7*(K*E2em.y*E2om.z - K*E2em.z*E2om.y + E2em.z*E2om.z*C_1)
		+ E2em.z*E2om.x*(Eep.x*C_6*(Eop.z*C_8 - Eop.y*K) + Eop.x*C_9*(Eep.y*K - Eep.z*C_4));
    ro = 
		(-E2em.x*E2om.y*K*(Eep.z*Ei.x*C_3 - Eep.x*Ei.z*C_4)
		+ E2em.x*E2om.z*(Eep.x*C_4*(Ei.z*C_5 - Ei.y*K) - Ei.x*C_3*(Eep.z*C_6 - Eep.y*K))
		- E2em.y*E2om.x*K*(Eep.x*Ei.z*C_6 - Eep.z*Ei.x*C_5)
		- Eep.x*Ei.x*K*C_2*(E2em.y*E2om.z - E2em.z*E2om.y)
		+ E2em.z*E2om.x*(Ei.x*C_5*(Eep.z*C_4 - Eep.y*K) - Eep.x*C_6*(Ei.z*C_3 - Ei.y*K))
		- E2em.z*E2om.z*(Eep.x*Ei.x*q2em*C_2 - Eep.x*Ei.x*q2om*C_2)
		+ E2em.x*E2om.x*C_1*(Eep.z*Ei.y*K - Eep.y*Ei.z*K - Eep.z*Ei.z*C_2))
		/ Nro;

	const double Nre = E2om.z*(E2em.x*q2em*(Eep.z*C_6 - Eep.y*K) + Eep.x*qep*(E2em.z*C_1 + E2em.y*K))
		- E2om.x*q2om*(E2em.z*(Eep.z*C_4 - Eep.y*K) + E2em.y*Eep.z*K)
		- E2om.y*K*(E2em.z*Eep.x*qep - E2em.x*Eep.z*q2em);
    re = abs(Nre) < 1e-8f ? .0f : (
		-(E2em.x*E2om.z*q2em*(Ei.z*C_5 - Ei.y*K - (Eop.y*K - Eop.z*C_9)*ro) 
		- E2em.z*E2om.x*q2om*(Ei.z*C_3 - Ei.y*K - (Eop.y*K - Eop.z*C_8)*ro) 
		+ E2em.z*E2om.z*C_1*(Ei.x*qi + Eop.x*qop*ro) 
		- K*(Ei.z + Eop.z*ro)*(E2em.y*E2om.x*q2om - E2em.x*E2om.y*q2em)
		- K*(Ei.x*qi + Eop.x*qop*ro)*(E2em.z*E2om.y - E2em.y*E2om.z))
		/ Nre);

    const double Nt = E2em.x*E2om.z - E2em.z*E2om.x;
    const double D_1 = Ei.z + Eep.z*re + Eop.z*ro;
    const double D_2 = Ei.x + Eep.x*re + Eop.x*ro;

    to = (E2em.x*D_1 - E2em.z*D_2) / Nt;
    te = (E2om.z*D_2 - E2om.x*D_1) / Nt;
}
void fresnel_aniso_aniso(float K, float eo, float eo2, float ee, float ee2, vec3 A, vec3 A2,
						 out float too, out float toe, out float teo, out float tee) {
	float ro,re;
	const float qo = -qo(eo,K);
	const float qe =  qe(eo,ee,A,K).y;
    const vec3 Eom = E_om(eo,A,K);
    const vec3 Eem = E_em(eo,ee,A,K);
	fresnel_aniso_aniso(K, qo, Eom, eo, eo2, ee, ee2, A, A2,
						ro, re, too, toe);
	fresnel_aniso_aniso(K, qe, Eem, eo, eo2, ee, ee2, A, A2,
						ro, re, teo, tee);
}
void fresnel_aniso_aniso(double K, double eo, double eo2, double ee, double ee2, dvec3 A, dvec3 A2,
						 out double too, out double toe, out double teo, out double tee) {
	double ro,re;
	const double qo = -qo(eo,K);
	const double qe =  qe(eo,ee,A,K).y;
    const dvec3 Eom = E_om(eo,A,K);
    const dvec3 Eem = E_em(eo,ee,A,K);
	fresnel_aniso_aniso(K, qo, Eom, eo, eo2, ee, ee2, A, A2,
						ro, re, too, toe);
	fresnel_aniso_aniso(K, qe, Eem, eo, eo2, ee, ee2, A, A2,
						ro, re, teo, tee);
}


// Precomputed functions go here
//%%%%% START
double h1r(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 		{ return 0; }
double h1i(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 		{ return 0; }
double h2r(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 		{ return 0; }
double h2i(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 		{ return 0; }
double Ig_ce1(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 		{ return 0; }
double Ig_ce2(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4)		{ return 1; }
dvec2  Ig1(const double K, const dvec2 Ga, const dvec2 Gb, const dvec2 erf0, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) { return dvec2(0); }
dvec2  Ig2(const double K, const dvec2 Ga, const dvec2 Gb, const dvec2 erf0, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) { return dvec2(0); }

double a(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 			{ return 1; }
double b(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 			{ return 0; }
double c(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 			{ return 0; }
double eo(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 			{ return 1.5; }
double ee(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 			{ return 1.5; }
double sigma_o(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4)		{ return 0; }
double sigma_e(const double K, const double h, const double c0, const double c1, const double c2, const double c3, const double c4) 	{ return 0; }
//%%%%% END

// Helpers
float h1r(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(h1r(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float h1i(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(h1i(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float h2r(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(h2r(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float h2i(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(h2i(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
double Ig_ce1(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 	{ return Ig_ce1(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4)); }
double Ig_ce2(const float K, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 	{ return Ig_ce2(double(K), double(h), double(c0), double(c1), double(c2), double(c3), double(c4)); }
vec2  Ig1(const float K, const dvec2 Ga, const dvec2 Gb, const dvec2 erf0, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) { return vec2(Ig1(double(K), Ga, Gb, erf0, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
vec2  Ig2(const float K, const dvec2 Ga, const dvec2 Gb, const dvec2 erf0, const float h, const float c0, const float c1, const float c2, const float c3, const float c4) { return vec2(Ig2(double(K), Ga, Gb, erf0, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float a(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 			{ return float(a(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float b(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 			{ return float(b(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float c(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 			{ return float(c(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float eo(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(eo(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float ee(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 		{ return float(ee(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float sigma_o(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 	{ return float(sigma_o(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }
float sigma_e(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) 	{ return float(sigma_e(0.lf, double(h), double(c0), double(c1), double(c2), double(c3), double(c4))); }

vec3 A(const float h, const float c0, const float c1, const float c2, const float c3, const float c4) {
	const float alpha = a(h,c0,c1,c2,c3,c4);
	const float beta  = b(h,c0,c1,c2,c3,c4);
	const float gamma = c(h,c0,c1,c2,c3,c4);
	return normalize(vec3(alpha, beta, gamma));
}
dvec3 A(const double h, const double c0, const double c1, const double c2, const double c3, const double c4) {
	const double alpha = a(0.lf,h,c0,c1,c2,c3,c4);
	const double beta  = b(0.lf,h,c0,c1,c2,c3,c4);
	const double gamma = c(0.lf,h,c0,c1,c2,c3,c4);
	return normalize(vec3(alpha, beta, gamma));
}

complex_dvec2 multilayered(dvec3 A0, double eo0, double ee0, double k, double K, double tau, complex_dvec2 Phi, const double c0, const double c1, const double c2, const double c3, const double c4) {
	double h = 0;
	const double dh = tau / double(layers);
	for (int l=0;l<layers;++l) {
		const dvec3 A1 = A(h+dh,c0,c1,c2,c3,c4);
		const double eo1 = eo(0.lf, h+dh,c0,c1,c2,c3,c4), ee1 = ee(0.lf, h+dh,c0,c1,c2,c3,c4);
		const double sigmao = sigma_o(0.lf, h,c0,c1,c2,c3,c4);
		const double sigmae = sigma_e(0.lf, h,c0,c1,c2,c3,c4);

		// Propagate
		const double qo  =-qo(eo0,K);
		const double qe  = qe(eo0,ee0,A0,K).y;
		const dvec3  Eom = E_om(eo0,A0,K);
		const dvec3  Eem = E_em(eo0,ee0,A0,K);
		const dvec3  Wom = dvec3(0, qo, -K);
		const dvec3  Wem = dvec3(0, qe, -K);
    	const dvec3  so  = normalize(Wom);// normalize(cross(Eom, cross(Wom, Eom)));
    	const dvec3  se  = normalize(Wem);// normalize(cross(Eem, cross(Wem, Eem)));
		const double vo  = 1.f / (-so.y);
		const double ve  = 1.f / (-se.y);

		const double ef = sqrt(sqr(qe) + sqr(K));

		// Attenuate
		const double att_o = exp(-max(0, vo*sigmao*dh));
		const double att_e = exp(-max(0, ve*sigmae*dh));
		//Phi.r *= vec2(att_o, att_e);
		//Phi.i *= vec2(att_o, att_e);

		const double Pe = (se.z*ve - so.z*vo)*K;
		const dmat2 R = dmat2(cos(k*dh*vo*eo0),0,0,cos(k*dh*(ve*ef + Pe)));
		const dmat2 I = dmat2(sin(k*dh*vo*eo0),0,0,sin(k*dh*(ve*ef + Pe)));
		const complex_dmat2 P = cdmat2(R, I);
		Phi = cdmat2_by_cdvec2(P, Phi);

		if (l!=layers-1) {
			// Fresnel coefficients
			double too, toe, teo, tee;
			fresnel_aniso_aniso(K, eo0, eo1, ee0, ee1, A0, A1, too, toe, teo, tee);

			const dmat2 T = dmat2(too, toe, teo, tee);
			Phi = dmat2_by_cdvec2(T, Phi);
		}

		// Next layer
		h = (l+1)/double(layers) * tau;
		A0 = A1;
		eo0 = eo1;
		ee0 = ee1;
	}

	return Phi;
}

// Computes the matrix exponentials of the solution
complex_vec2 matrix_exp_solution(complex_vec2 oe, float k, float h1r, float h1i, float h2r, float h2i, float g_exp_i, vec2 Ig1, vec2 Ig2, bool gaussian) {
	// Avoid overflows with numerically unstable values
	const float g3r = clamp((h1r - h2r)/2.0, -25.f, 25.f);
	const float fr  = clamp((h1r + h2r)/2.0, -25.f, 25.f);
	const float g3i = (h1i - h2i)/2.0;
    const float mui = 2*g3i - g_exp_i;
	const float C   = gaussian ? sqrt(2*pi) : 1;

	const vec2 g1 = C * exp( 2*g3r) * product(cexpi( k*mui), Ig1);
	const vec2 g2 = C * exp(-2*g3r) * product(cexpi(-k*mui), Ig2);

	const float efr = exp(fr);
	const float eg3 = exp(g3r);

	const vec2 S00  = efr*eg3 * product(vec2(1,0) + product(g1,g2), cexpi(k*h1i));
	const vec2 S10  = efr*eg3 * product(g2, cexpi(k*h1i));
	const vec2 S01  = efr/eg3 * product(g1, cexpi(k*h2i));
	const vec2 S11  = efr/eg3 * cexpi(k*h2i);

	const complex_mat2 S = cmat2(mat2(S00.x,S10.x,S01.x,S11.x), mat2(S00.y,S10.y,S01.y,S11.y));

	complex_vec2 ret = cmat2_by_cvec2(S, oe);
	// Hack to clean up numerical instabillity (only happens when approximation starts to break down, relevant for
	// pi-wall-defect profile under moderate to strong electric field only).
	const float mag_ret = sqrt(dot(mag2(ret),vec2(1)));
	if (mag_ret > 1) { ret.r /= mag_ret; ret.i /= mag_ret; }

	return ret;
}

void main() {
	const int render_mode = view_mode & 255;
	const bool screenshot_mode = (view_mode & (1<<8)) != 0;

	// Normal and exitant vectors
	const vec3 n = normalize(cross(dFdx(wposition),dFdy(wposition)));
	const vec3 o = normalize(-position * mat3(ubo.invModelView));

	// Surface lighting
	const float roughness = 0.4;
	const vec3 L = normalize(lightPosition - wposition);
	const vec3 Lo = BRDF(L, o, n, 0, roughness) * (render_mode != 0 ? .5f : 1.f);
	vec3 color = materialcolor() * 0.005f;
	color += Lo * (screenshot_mode ? 1e-5f : 0.25f);

	const vec3 p_to_l = wposition.xyz - ubo.lightPos.xyz;
	float d = length(p_to_l.xz);
	const float R0 = 1.0f;
	const float R1 = 1.3f;
	const float incident_energy = 1-clamp((d-R0)/(R1-R0), .0f, 1.f);

	// Polarization filter
	const vec2 pfilter = render_mode == 0 ? vec2(1,1) : render_mode == 1 ? vec2(1,1e-3f) : vec2(1e-3f,1);

	// Incident light
	const float s_in	 = sqrt(incident_energy * (1-params.polarization));
	const float p_in	 = sqrt(incident_energy * (params.polarization));
	const vec2 LT		 = d>0 ? p_to_l.xz/d : vec2(1,0);
	const mat2 toLocal   = mat2(LT.x, LT.y, -LT.y, LT.x);
	const mat2 fromLocal = mat2(LT.x, -LT.y, LT.y, LT.x);

	complex_vec2 Phi_in = cvec2(vec2(s_in,0), vec2(p_in,0));

	// Incidence vector at local frame
	const vec3 I = normalize(ubo.lightPos.xyz - wposition);
	float cos_theta = -I.y;
	float sin_theta = sqrt(1-sqr(cos_theta));
	float incident_sin_theta = sin_theta;

	float h = params.tau;			// Slab thickness
	float wavelength_scale = 1e-6f;	// mm
	// Refractive index of surrounding media
	float ei = profile_type==profile_slab ? 1.0 : 1.42;
	// Control constants
	float c0 = .0f, c1 = 1.f, c2 = .0f, c3 = .0f, c4 = .0f;
	c0 = params.c0;
	c1 = params.c1;
	c2 = params.c2;
	
	// Droplet/shells specifics
	float droplet_rad = 0;
	float start_angle, end_angle;
	float sin_theta_inner, inner_rad;
	if (profile_type != profile_slab && profile_type != profile_heat) {
		const float droplet_rescale = 1.f/3.f;
		d *= droplet_rescale;
		// Additional rescale for large droplets/shells
		if (profile_type == profile_droplet || profile_type == profile_shell)
			d /= 3.8;

		float maxR = 1;
		if (profile_type >= profile_microdroplet && profile_type <= profile_droplet_pi_wall) {
			// micron sized droplets
			maxR = 20;
			wavelength_scale = 1e-3f / maxR;
		}
		droplet_rad = params.tau/2 / maxR;
		sin_theta = d/droplet_rad;

		// Shells specifics
		if (profile_type == profile_shell) {
			const float shell_thickness = c2*.01f;		// c2 in percentage
			inner_rad = droplet_rad*(1-shell_thickness);
			sin_theta_inner = min(d/inner_rad, 1);
		}

		incident_sin_theta = sin_theta;
		if (incident_sin_theta >= 0.99) incident_sin_theta = 1;		// Hack to avoid a numerically unstable region at the very edge. A more accurate profile is required.
		
		if (profile_type != profile_shell) {
			// Approximate refraction
			const float eta_ev = (2*eo(0,0,0,0,0,0)+ee(0,0,0,0,0,0))/3.f;
			sin_theta *= ei/eta_ev;
		}
		
		sin_theta = min(sin_theta, 1);
		cos_theta = sqrt(1-sqr(sin_theta));
		h = cos_theta * 2*droplet_rad;

		start_angle = pi * (1-cos_theta)/2;
		end_angle   = pi - start_angle;
		// axial droplet with disclination defect under electric field
		if (profile_type == profile_droplet_axial) {
			end_angle   = -start_angle;
			// c2 controls the simulation of the opening of the defect into a disclination ring
			h /= mix(1,3,c2);
		}
		
		c3 = start_angle;
		c4 = (end_angle - start_angle) / h;
	}
	const float cos_theta_inner = sqrt(1-sqr(sin_theta_inner));

	// Angle, to re-align the optic axis for slabs
	float zeta = -atan(p_to_l.x, p_to_l.z);
	if (profile_type == profile_slab || profile_type == profile_heat)
		c4 = zeta + pi*c2;

	// Heat-conduction in a liquid crystal plate
	if (profile_type == profile_heat) {
		c2 = c1;
		c3 = h;		// Slab thickness (used for normalization)
	}

	// Incidence parameter
	float K = ei * incident_sin_theta;

	// Material constants at interfaces
	const vec3  A0  =  A(0,c0,c1,c2,c3,c4);
	const float eo0 = eo(0,c0,c1,c2,c3,c4);
	const float ee0 = ee(0,c0,c1,c2,c3,c4);
	const vec3  Ah  =  A(h,c0,c1,c2,c3,c4);
	const float eoh = eo(h,c0,c1,c2,c3,c4);
	const float eeh = ee(h,c0,c1,c2,c3,c4);

	// Refraction into/out of the slab
	float rss, rsp, tso, tse, rps, rpp, tpo, tpe;
	fresnel_iso_aniso(cos_theta, sin_theta, A0, ei, eo0, ee0, rss, rsp, tso, tse, rps, rpp, tpo, tpe);
	float roo, roe, tos, top, reo, ree, tes, tep;
	fresnel_aniso_iso(cos_theta, sin_theta, Ah, eoh, eeh, ei, roo, roe, tos, top, reo, ree, tes, tep);

	const mat2 T_in = mat2(tso, tse, tpo, tpe) * toLocal;
	const mat2 T_out = fromLocal * mat2(tos, top, tes, tep);

	// Ignore sides
	// And area outside droplet
	if (n.y<-0.99 && incident_energy>.0f && 
		(
			 profile_type == profile_slab ||
			 profile_type == profile_heat ||
			(profile_type >= profile_droplet && profile_type <= profile_shell && sin_theta<1 && sin_theta>=0 && incident_sin_theta<1)
		)) {
		complex_vec2 Phi_0 = mat2_by_cvec2(T_in, Phi_in);
		if (profile_type == profile_droplet_pi_wall) Phi_0 = mat2_by_cvec2(T_in, cvflip(Phi_in));
		
		const int wavelength_quadrature_steps = white_light > 0 ? 16 : 1;
		const float wavelength_start = white_light > 0 ? 400.f : ubo.wavelength;
		const float wavelength_end   = white_light > 0 ? 700.f : ubo.wavelength;

		complex_vec2 Phi_h[wavelength_quadrature_steps];
		for (int WL=0; WL<wavelength_quadrature_steps; ++WL)
			Phi_h[WL] = Phi_0;

		int Chunks = 1;
		if (profile_type != profile_slab && profile_type != profile_heat)
			Chunks = 4;
		if (profile_type == profile_droplet_axial)
			Chunks = screenshot_mode?4:2;
		if (profile_type == profile_droplet_pi_wall)
			Chunks = 2;
		if (profile_type == profile_shell)
			Chunks = sin_theta_inner==1?4:2;
		
		float H = h/float(Chunks);
		if (profile_type == profile_shell && sin_theta_inner<1) {
			// Shells are split into 2 groups of chunks, upper and lower. Chunks assumed to be even.
			const float h_inner = cos_theta_inner * 2*inner_rad;
			H = (h-h_inner) / float(Chunks);
		}

		// Propagation in the slab
		for (int C=0;C<Chunks;++C) {
			float C2 = profile_type == profile_heat ? c2 : 1;
			float C3 = c3;
			float C4 = c4;
			
			if (profile_type != profile_slab && profile_type != profile_heat)
				C3 = c3 + H*float(C)*c4;
			if (profile_type == profile_droplet_pi_wall) {
				// Emulate the collapsing walls
				const int walls_chunks = 2;
				const float collapse_travel = 0.3333f;
				const float collapse_angle_frac = collapse_travel * 4e-1f;
				if (C < walls_chunks || C >= Chunks-walls_chunks) {
					H = h/(walls_chunks) * collapse_travel;
					C2 = 1 - collapse_travel*1e+1f;
					C4 = c4 * (collapse_angle_frac / collapse_travel);
					C3 = c3 + H*float(C)*C4;
					if (C >= Chunks-walls_chunks)
						C3 = c3 + (h-float(Chunks-C)*H)*C4;
				}
				else {
					H = h/float(Chunks-2*walls_chunks) * (1-collapse_travel);
					C2 = 1 - collapse_travel*1e+1f;
					C4 = c4 * (1-collapse_angle_frac) / (1-collapse_travel);
					C3 = c3 + (end_angle-start_angle)*collapse_angle_frac/2 + H*float(C-walls_chunks)*C4;
				}
			}
			if (profile_type == profile_shell && sin_theta_inner < 1) {
				if (C >= Chunks/2)
					C3 = c3 + (h-float(Chunks-C)*H)*c4;

				if (C == Chunks/2) {
					// Refract into inner bubble, and back into the shell.
					const float shell_inner_eta = ei;
					const float C3_prev = c3 + H*float(C-1)*c4;
					const vec3  A1  = A(H,c0,c1,c2,C3_prev,C4);
					const vec3  A2  = A(0,c0,c1,c2,C3,C4);
					const float eo  = eo0, ee = ee0;
					float roo, roe, tos, top, reo, ree, tes, tep, rss, rsp, tso, tse, rps, rpp, tpo, tpe;
					fresnel_aniso_iso(cos_theta_inner, sin_theta_inner, A1, eo, ee, shell_inner_eta, roo, roe, tos, top, reo, ree, tes, tep);
					fresnel_iso_aniso(cos_theta_inner, sin_theta_inner, A2, shell_inner_eta, eo, ee, rss, rsp, tso, tse, rps, rpp, tpo, tpe);
					const mat2 T = mat2(tso, tse, tpo, tpe) * mat2(tos, top, tes, tep);
						
					for (int WL=0; WL<wavelength_quadrature_steps; ++WL)
						Phi_h[WL] = mat2_by_cvec2(T, Phi_h[WL]);
				}
			}
			
			// Wavelength-independant quantities
			const float  H1r     =    h1r(K,H,c0,c1,C2,C3,C4);
			const float  H1i     =    h1i(K,H,c0,c1,C2,C3,C4);
			const float  H2r     =    h2r(K,H,c0,c1,C2,C3,C4);
			const float  H2i     =    h2i(K,H,c0,c1,C2,C3,C4);
			const double ce1     = Ig_ce1(K,H,c0,c1,C2,C3,C4);
			const double ce2     = Ig_ce2(K,H,c0,c1,C2,C3,C4);

			dvec2 Ig_Ga, Ig_Gb;
			float exp_i = 0;
			if (ce2 != 0) {
				const dvec2 sqrt_ce2 = csqrtd(ce2);
				Ig_Ga    = 1.lf/sqrt(2.lf) * sqrt_ce2;
				Ig_Gb    = 1.lf/sqrt(2.lf) * cddiv(ce1, sqrt_ce2);
				exp_i    = .5f * float(sqr(ce1)/ce2);
			}
				
			for (int WL=0; WL<wavelength_quadrature_steps; ++WL) {
				// wavenumber
				const float wavelength = wavelength_start + (wavelength_end - wavelength_start) / float(wavelength_quadrature_steps) * (float(WL)+.5f);
				const float lambda = wavelength * wavelength_scale;
				const float k = 2*pi/lambda;

				if (layers > 0) {
					// Render a multi-layered structure of homogeneous anisotrpic layers
					Phi_h[WL] = cvec2(multilayered(A0, eo0, ee0, k, K, H, cdvec2(Phi_h[WL]),
												   c0,c1,C2,C3,C4));
				}
				else {
					// Render with our solution to the radiative transfer in an inhomogeneous anisotropic slab
					dvec2 Ga1, Gb1, Ga2, Gb2, Ig1_erf0, Ig2_erf0;
					bool gaussian = true;
					if (ce2 != 0) {
						Ga2 = sqrt(k) * cdprod(dvec2(1,-1), Ig_Ga);
						Gb2 = sqrt(k) * cdprod(dvec2(1,-1), Ig_Gb);
						Ig2_erf0 = Ig_erf0(Ga2,Gb2,H);

						Ga1 = cdprod(dvec2(0,1), Ga2);	// i*Ga
						Gb1 = cdprod(dvec2(0,1), Gb2);	// i*Gb
						Ig1_erf0 = Ig_erf0(Ga1,Gb1,H);
					}
					else {
						Ga1 =  dvec2(0,1) * k * ce1;
						Ga2 = -dvec2(0,1) * k * ce1;
						Gb1 = Gb2 = Ig1_erf0 = Ig2_erf0 = dvec2(0);
						gaussian = false;
					}

					const vec2 I_g1 = Ig1(K,Ga1,Gb1,Ig1_erf0,H,c0,c1,C2,C3,C4);
					const vec2 I_g2 = Ig2(K,Ga2,Gb2,Ig2_erf0,H,c0,c1,C2,C3,C4);

					Phi_h[WL] = matrix_exp_solution(Phi_h[WL], k, H1r, H1i, H2r, H2i, exp_i, I_g1, I_g2, gaussian);
				}
			}
		}

		for (int WL=0; WL<wavelength_quadrature_steps; ++WL) {
			const float wavelength = wavelength_start + (wavelength_end - wavelength_start) / float(wavelength_quadrature_steps) * (float(WL)+.5f);
			
			if (profile_type == profile_droplet_pi_wall) Phi_h[WL] = cvflip(Phi_h[WL]);
			const complex_vec2 Phi_out = mat2_by_cvec2(T_out, Phi_h[WL]);		// Refract
			const float intensity = dot(mag2(Phi_out),pfilter);					// Polarization filter

			if (!screenshot_mode || (!isinf(intensity) && !isnan(intensity))) {
				const vec3 XYZ = sqr(intensity) * XYZ_1931(wavelength) / float(wavelength_quadrature_steps);
				color += XYZtoRGB(XYZ);
			}
		}
	}
	else {
		const vec3 lightColor = XYZtoRGB(white_light>0 ? white_point() : XYZ_1931(ubo.wavelength));
		color += sqr(dot(sqr(vec2(s_in,p_in)), pfilter)) * lightColor;
	}
	
	outFragColor = vec4(color, 1);
}
