/*
*
*	Example implementation for "Accurate Rendering of Liquid-Crystals and Inhomogeneous Optically Anisotropic Media"
*
* This code is licensed under the MIT license (MIT) (http://opensource.org/licenses/MIT)
*/


#version 450

#extension GL_ARB_separate_shader_objects : enable
#extension GL_ARB_shading_language_420pack : enable

layout (binding = 0) uniform restrict UBO {
	mat4 projection;
	mat4 modelView;
	mat4 invModelView;
	vec4 cameraPosition;
	vec4 lightPos;
	vec4 lightRadii;
	float wavelength;
	int white_light;
	int profile_type;
} ubo;
layout (binding = 1) uniform restrict Params {
	float t;
	float exposure;
	float tau;
	float polarization;
} params;

layout (location = 0) out vec4 outFragColor;

layout (location = 0) in vec3 color;
layout (location = 1) in vec3 position;
layout (location = 2) in vec3 wposition;

const float pi = 3.1415f;


float sqr(float x) { return x*x; }
// Adobe XYZ to RGB
// (used as an example)
vec3 XYZtoRGB(vec3 XYZ) {
	return XYZ * mat3( 2.0413690, -0.5649464, -0.3446944,
					  -0.9692660,  1.8760108,  0.0415560,
					   0.0134474, -0.1183897,  1.0154096);
}
vec3 white_point() { return vec3(0.9504, 1.0000, 1.0888); }

// Inputs:  Wavelength in nanometers
// From "Simple Analytic Approximations to the CIE XYZ Color Matching Functions"
float X_1931(float wave) {
	const float t1 = (wave-442.0f) * ((wave<442.0f)?0.0624f:0.0374f);
	const float t2 = (wave-599.8f) * ((wave<599.8f)?0.0264f:0.0323f);
	const float t3 = (wave-501.1f) * ((wave<501.1f)?0.0490f:0.0382f);
	return 0.362f * exp(-0.5f * sqr(t1)) + 1.056f * exp(-0.5f * sqr(t2)) - 0.065f * exp(-0.5f * sqr(t3));
}
float Y_1931(float wave) {
	const float t1 = (wave-568.8f) * ((wave<568.8f)?0.0213f:0.0247f);
	const float t2 = (wave-530.9f) * ((wave<530.9f)?0.0613f:0.0322f);
	return 0.821f * exp(-0.5f * sqr(t1)) + 0.286f * exp(-0.5f * sqr(t2));
}
float Z_1931(float wave) {
	const float t1 = (wave-437.0f) * ((wave<437.0f)?0.0845f:0.0278f);
	const float t2 = (wave-459.0f) * ((wave<459.0f)?0.0385f:0.0725f);
	return 1.217f * exp(-0.5f * sqr(t1)) + 0.681f * exp(-0.5f * sqr(t2));
}


//	Classic Perlin 3D Noise 
//	by Stefan Gustavson
vec4 permute(vec4 x){return mod(((x*34.0)+1.0)*x, 289.0);}
vec4 taylorInvSqrt(vec4 r){return 1.79284291400159 - 0.85373472095314 * r;}
vec3 fade(vec3 t) {return t*t*t*(t*(t*6.0-15.0)+10.0);}
float cnoise(vec3 P){
	vec3 Pi0 = floor(P); // Integer part for indexing
	vec3 Pi1 = Pi0 + vec3(1.0); // Integer part + 1
	Pi0 = mod(Pi0, 289.0);
	Pi1 = mod(Pi1, 289.0);
	vec3 Pf0 = fract(P); // Fractional part for interpolation
	vec3 Pf1 = Pf0 - vec3(1.0); // Fractional part - 1.0
	vec4 ix = vec4(Pi0.x, Pi1.x, Pi0.x, Pi1.x);
	vec4 iy = vec4(Pi0.yy, Pi1.yy);
	vec4 iz0 = Pi0.zzzz;
	vec4 iz1 = Pi1.zzzz;

	vec4 ixy = permute(permute(ix) + iy);
	vec4 ixy0 = permute(ixy + iz0);
	vec4 ixy1 = permute(ixy + iz1);

	vec4 gx0 = ixy0 / 7.0;
	vec4 gy0 = fract(floor(gx0) / 7.0) - 0.5;
	gx0 = fract(gx0);
	vec4 gz0 = vec4(0.5) - abs(gx0) - abs(gy0);
	vec4 sz0 = step(gz0, vec4(0.0));
	gx0 -= sz0 * (step(0.0, gx0) - 0.5);
	gy0 -= sz0 * (step(0.0, gy0) - 0.5);

	vec4 gx1 = ixy1 / 7.0;
	vec4 gy1 = fract(floor(gx1) / 7.0) - 0.5;
	gx1 = fract(gx1);
	vec4 gz1 = vec4(0.5) - abs(gx1) - abs(gy1);
	vec4 sz1 = step(gz1, vec4(0.0));
	gx1 -= sz1 * (step(0.0, gx1) - 0.5);
	gy1 -= sz1 * (step(0.0, gy1) - 0.5);

	vec3 g000 = vec3(gx0.x,gy0.x,gz0.x);
	vec3 g100 = vec3(gx0.y,gy0.y,gz0.y);
	vec3 g010 = vec3(gx0.z,gy0.z,gz0.z);
	vec3 g110 = vec3(gx0.w,gy0.w,gz0.w);
	vec3 g001 = vec3(gx1.x,gy1.x,gz1.x);
	vec3 g101 = vec3(gx1.y,gy1.y,gz1.y);
	vec3 g011 = vec3(gx1.z,gy1.z,gz1.z);
	vec3 g111 = vec3(gx1.w,gy1.w,gz1.w);

	vec4 norm0 = taylorInvSqrt(vec4(dot(g000, g000), dot(g010, g010), dot(g100, g100), dot(g110, g110)));
	g000 *= norm0.x;
	g010 *= norm0.y;
	g100 *= norm0.z;
	g110 *= norm0.w;
	vec4 norm1 = taylorInvSqrt(vec4(dot(g001, g001), dot(g011, g011), dot(g101, g101), dot(g111, g111)));
	g001 *= norm1.x;
	g011 *= norm1.y;
	g101 *= norm1.z;
	g111 *= norm1.w;

	float n000 = dot(g000, Pf0);
	float n100 = dot(g100, vec3(Pf1.x, Pf0.yz));
	float n010 = dot(g010, vec3(Pf0.x, Pf1.y, Pf0.z));
	float n110 = dot(g110, vec3(Pf1.xy, Pf0.z));
	float n001 = dot(g001, vec3(Pf0.xy, Pf1.z));
	float n101 = dot(g101, vec3(Pf1.x, Pf0.y, Pf1.z));
	float n011 = dot(g011, vec3(Pf0.x, Pf1.yz));
	float n111 = dot(g111, Pf1);

	vec3 fade_xyz = fade(Pf0);
	vec4 n_z = mix(vec4(n000, n100, n010, n110), vec4(n001, n101, n011, n111), fade_xyz.z);
	vec2 n_yz = mix(n_z.xy, n_z.zw, fade_xyz.y);
	float n_xyz = mix(n_yz.x, n_yz.y, fade_xyz.x); 
	return 2.2 * n_xyz;
}


void main() {
	float alpha=1;

	const float wavelength = ubo.wavelength;
	const vec3 lightColor = XYZtoRGB(ubo.white_light>0 ? white_point() : vec3(X_1931(wavelength), Y_1931(wavelength), Z_1931(wavelength)));

	// Height at which to start fading out
	const float max_l = 2.7f;

	const vec4 L  = ubo.modelView * vec4(ubo.lightPos.xyz,1);
	const vec3 U  = normalize(position.xyz);
	const vec3 d  = -L.xyz;
	const vec3 D  = mat3(ubo.invModelView) * vec3(0,1,0);

	vec3 pos1 = vec3(0), pos2 = vec3(0);
	// Slab?
	if (ubo.profile_type == 0 || ubo.profile_type == 7) {
		// Ray-cone intersection test
		const float c  = dot(vec3(0,1,0),normalize(vec3(0,1,ubo.lightRadii.y)));
		const mat3  M  = outerProduct(D,D) - c*c*mat3(1);
		const float c0 = dot(d*M, d);
		const float c1 = dot(U*M, d);
		const float c2 = dot(U*M, U);
		const float det = c1*c1-c0*c2;
		if (det<=0) {
			// No intersection
			outFragColor = vec4(0);
			return;
		}
	
		float t1 = (-c1 + sqrt(det))/c2;
		float t2 = (-c1 - sqrt(det))/c2;
		if (dot(t1*U-L.xyz,D)<0 || dot(t2*U-L.xyz,D)<0) {
			// Discard shadow cone
			outFragColor = vec4(0);
			return;
		}
		if (t1<0 || t2<0) {
			t2 = max(t1,t2);
			t1 = t2+100;
		}
		pos1 = t1*U;
		pos2 = t2*U;
	}
	else {
		// Ray-cylinder intersection test
		const float r = 1.3f;
		const vec3  O = transpose(ubo.invModelView)[3].xyz;
		const vec3  V = normalize(wposition.xyz - O);
		
		const float a = 2*V.x*V.x + 2*V.z*V.z;
		const float b = -2*O.x*V.x - 2*O.z*V.z;
		const float c = O.x*O.x + O.z*O.z - r*r;
		const float det = b*b - 2*a*c;
		if (det<=0) {
			// No intersection
			outFragColor = vec4(0);
			return;
		}

		const float t1 = (b - sqrt(det)) / a;
		const float t2 = (b + sqrt(det)) / a;
			
		pos1 = t1*U;
		pos2 = t2*U;

		alpha *= 0.075f;
		alpha *= clamp(abs(t1-t2)*1.5/r,.0f,1.f);
	}

	const float d1 = dot(pos1+d,D);
	const float d2 = dot(pos2+d,D);
		
	alpha *= min(1,length(pos1-pos2)/4.f);
	alpha *= smoothstep(0,1,1-d1/max_l) * smoothstep(0,1,1-d2/max_l);

	if (alpha>.0f)
		alpha *= 0.4f + 0.6f*clamp(cnoise(vec3(gl_FragCoord.xy/150,params.t/2.f)),0,1);

	outFragColor = vec4(lightColor.rgb, alpha);
}
