/*
*
*	Example implementation for "Accurate Rendering of Liquid-Crystals and Inhomogeneous Optically Anisotropic Media"
*	 by Shlomi Steinberg
*	 http://ssteinberg.xyz
*
*
*    This program is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <vector>
#include <random>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <sstream>
#include <filesystem>
#include <optional>
#include <atomic>
#include <future>

#define GLM_FORCE_RADIANS
#define GLM_FORCE_DEPTH_ZERO_TO_ONE
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>

// color-util expects M_PI to be defined, for reasons
#undef M_PI
#define M_PI glm::pi<double>()
#include <color-util/XYZ_to_Lab.hpp>
#include <color-util/CIEDE2000.hpp>
#undef M_PI

#include <vulkan/vulkan.h>
#include "base/vulkanexamplebase.h"
#include "base/VulkanBuffer.hpp"
#include "base/VulkanTexture.hpp"
#include "base/VulkanModel.hpp"

#ifdef _linux
#include <png.h>
#elif defined(_WIN32)
#include <FreeImage.h>
#else
#error Unkown platform
#endif

#include "profiles_manager.hpp"


#define VERTEX_BUFFER_BIND_ID 0
#define ENABLE_VALIDATION false


#ifdef _linux
inline bool write_2d(const char* file_name, const uint8_t *image_data, int components, int width, int height) {
	FILE *fp = fopen(file_name, "wb");
	if (!fp) {
		std::cerr << file_name << " can't be opened for writing" << std::endl;
		return false;
	}

	png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);
	if (!png) {
		std::cerr << file_name << " png_create_write_struct failed" << std::endl;
		fclose(fp);
		return false;
	}

	png_infop info = png_create_info_struct(png);
	if (!info) {
		std::cerr << file_name << " png_create_info_struct failed" << std::endl;
		fclose(fp);
		return false;
	}

	if (setjmp(png_jmpbuf(png))) {
		std::cerr << file_name << " png_jmpbuf failed" << std::endl;
		fclose(fp);
		return false;
	}

	png_byte ** const row_pointers = reinterpret_cast<png_byte **>(malloc(height * sizeof(png_byte *)));
	if (row_pointers == nullptr) {
		std::cerr << file_name << " could not allocate memory for PNG row pointers" << std::endl;
		fclose(fp);
		return false;
	}

	// set the individual row_pointers to point at the correct offsets of image_data
	// To maintain compatibility png_write_image requests a non-const double pointer, hack the const away...
	for (int i = 0; i < height; i++)
		row_pointers[i] = const_cast<png_byte*>(image_data + i * width * components);

	png_init_io(png, fp);

	int color_type;
	switch (components) {
	case 1: color_type = PNG_COLOR_TYPE_GRAY; break;
	case 3: color_type = PNG_COLOR_TYPE_RGB; break;
	case 4: color_type = PNG_COLOR_TYPE_RGBA; break;
	default: {
		std::cerr << "Invalid component count" << std::endl;
		return false;
	}
	}
	png_set_IHDR(png,
		info,
		width, height,
		8,
		color_type,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png, info);

	png_write_image(png, row_pointers);
	png_write_end(png, nullptr);

	free(row_pointers);
	fclose(fp);

	return true;
}
#elif defined(_WIN32)
inline bool write_2d(const char* file_name, const uint8_t *image_data, int components, int width, int height) {
	FIBITMAP *img = FreeImage_ConvertFromRawBits(const_cast<BYTE*>(image_data), width, height, components * width, components * 8, FI_RGBA_RED_MASK, FI_RGBA_GREEN_MASK, FI_RGBA_BLUE_MASK, true);
	const auto ret = FreeImage_Save(FIF_PNG, img, file_name);
	FreeImage_Unload(img);

	return ret;
}
#endif

class VulkanExample : public VulkanExampleBase
{
public:
	struct {
		VkPipelineVertexInputStateCreateInfo inputState;
		std::vector<VkVertexInputBindingDescription> bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
	} vertices;

	// Vertex layout for the models
	vks::VertexLayout vertexLayout = vks::VertexLayout({
		vks::VERTEX_COMPONENT_POSITION,
		vks::VERTEX_COMPONENT_NORMAL,
		vks::VERTEX_COMPONENT_UV,
		vks::VERTEX_COMPONENT_COLOR,
	});

	struct {
		vks::Model object;
		vks::Model cylinder;
		vks::Model sphere;
	} models;

	struct {
	} textures;

	struct {
		vks::Buffer object, params;
	} uniformBuffers;

	struct UBOVS {
		glm::mat4 projection;
		glm::mat4 modelView;
		glm::mat4 invModelView;
		glm::vec4 cameraPosition;
		glm::vec4 lightPosition = { 0,-3.6,0,0 };
		glm::vec4 lightRadii = { 0.3,0.425,0,0 };
		float wavelength = 590;
		int32_t white_light = 0;
		int32_t profile_type = 0;
	} uboVS;
	struct UBOParams {
		float time = .0f;
		float exposure = 1.f;
		float d = .4f;
		float polarization = 1.f;
		// Constrol constants
		float c0 = .0f, c1 = .0f, c2 = .0f, c3 = .0f, c4 = .0f;
	} uboParams;

	int32_t view_mode = 1;
	int32_t white_light = 0;
	int32_t layers = 0;

	bool animate_wavelength{ false };
	float animate_wavelength_key{ .0f };
	std::chrono::steady_clock::time_point animate_wavelength_timestamp;

	bool render_props = true;

	const std::chrono::steady_clock::time_point start_time = std::chrono::steady_clock::now();

	struct screenshot_context_t {
		std::string name;

		glm::vec3 prev_camera_pos, prev_camera_rot;
		bool update_camera{ false };
		
		int32_t prev_view_mode;
		int32_t prev_white_light;
		bool prev_render_props;
	};
	std::optional<screenshot_context_t> screenshot_context;

	static constexpr glm::vec3 plate_scale   = glm::vec3(2.3f, 1.f,  2.3f);
	static constexpr glm::vec3 plate_offset  = glm::vec3(.0f, -1.8f, .0f);
	static constexpr glm::vec3 volume_scale  = glm::vec3(5.f,  50.f, 5.f);
	static constexpr glm::vec3 volume_offset = glm::vec3(.0f, -2.8f, .0f);

	struct {
		VkPipeline scene;
		VkPipeline composition;
		VkPipeline object;
		VkPipeline render;
		VkPipeline volume;
	} pipelines;
	struct {
		const uint32_t count = 5;
		VkDescriptorSet model;
		VkDescriptorSet floor;
		VkDescriptorSet scene;
		VkDescriptorSet composition;
	} descriptorSets;
	struct {
		VkPipelineLayout scene;
		VkPipelineLayout composition;
		VkPipelineLayout object;
		VkPipelineLayout render;
	} pipelineLayouts;
	struct {
		VkDescriptorSetLayout scene;
		VkDescriptorSetLayout composition;
	} descriptorSetLayouts;

	struct FrameBufferAttachment {
		VkImage image;
		VkDeviceMemory mem;
		VkImageView view;
		VkFormat format;
		void destroy(VkDevice device)
		{
			vkDestroyImageView(device, view, nullptr);
			vkDestroyImage(device, image, nullptr);
			vkFreeMemory(device, mem, nullptr);
		}
	};
	struct FrameBuffer {
		int32_t width, height;
		VkFramebuffer frameBuffer;
		FrameBufferAttachment color;
		FrameBufferAttachment depth;
		VkRenderPass renderPass;
		VkSampler sampler;
		VkCommandBuffer cmdBuffer = VK_NULL_HANDLE;
		VkSemaphore semaphore = VK_NULL_HANDLE;
	} offscreen;


	profiles_manager profiles;
	std::vector<std::string> profile_names;

	using profile_t = profiles_manager::profile_t;
	using profile_type_t = profiles_manager::profile_type_t;

	void reload_profiles() {
		profiles.reload_profiles();

		profile_names.reserve(profiles.profiles.size()+1);
		profile_names.clear();
		profile_names.emplace_back();
		for (const auto &p : profiles.profiles)
			profile_names.emplace_back(p.name);
	}


	/***
	 * 	Materials
	 */
	struct material_t {
		// Profile and thickness
		profile_t profile;
		float tau;
	} selected_material{}, current_material{};
	profile_type_t current_type{ profile_type_t::slab };

	static std::string generate_glsl_for_material(const material_t &mat) {
		return  mat.profile.Ig1 + "\n" + 
				mat.profile.Ig2 + "\n" + 
				mat.profile.Ig_ce1 + "\n" + 
				mat.profile.Ig_ce2 + "\n" + 
				mat.profile.h1r + "\n" + 
				mat.profile.h1i + "\n" + 
				mat.profile.h2r + "\n" + 
				mat.profile.h2i + "\n" + 
				mat.profile.a + "\n" + 
				mat.profile.b + "\n" + 
				mat.profile.c + "\n" + 
				mat.profile.eo + "\n" + 
				mat.profile.ee + "\n" + 
				mat.profile.sigma_o + "\n" + 
				mat.profile.sigma_e + "\n\n";
	}
	
	bool compile_material(const material_t &mat) {
		std::filesystem::create_directory("../tmp");
		std::filesystem::path dst = "../tmp/shader.frag";

		{
			const std::filesystem::path shader_path = spirv_src[mat.profile.version];
			std::ifstream f(shader_path);
			if (!f) {
				std::cerr << "Source shader '" << shader_path.string() << "' not found!" << std::endl;
				return false;
			}
			std::string shader;
			std::string line;
			bool input_injected = false;
			while (std::getline(f, line)) {
				if (line == "//%%%%% START") {
					while (std::getline(f, line) && line!="//%%%%% END") {}
					input_injected=true;
					shader += generate_glsl_for_material(mat);
				}
				else
					shader += line+"\n";
			}
			if (!input_injected) {
				std::cerr << "Failed to find '//%%%%% START' and '//%%%%% END' pair for code injection into shader. Source shader ('" << shader_path.string() << "') modified?" << std::endl;
				return false;
			}

			std::ofstream(dst) << shader;
		}

		// Compile SPIR-v
		const std::string glsllang_cmd = "glslangValidator -V \"" + dst.string() + "\" -o \"" + spirv_dst.string() + "\"";
#ifdef _linux
		std::cout << "Executing " << glsllang_cmd << std::endl;
		return system(glsllang_cmd.c_str()) == 0;
#elif defined (_WIN32)
		STARTUPINFOA si;
		PROCESS_INFORMATION pi;
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);
		si.lpTitle = "glslangValidator";
		si.wShowWindow = SW_NORMAL;
		si.dwFlags |= STARTF_USESHOWWINDOW;
		ZeroMemory(&pi, sizeof(pi));

		const auto ret = CreateProcess(nullptr, (LPSTR)glsllang_cmd.c_str(), nullptr, nullptr, true, 0, nullptr, nullptr, &si, &pi);
		if (!!ret)
			WaitForSingleObject(pi.hProcess, INFINITE);
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);

		return !!ret;
#endif
	}

	/***
	 * 	Application
	 */

	VulkanExample() : VulkanExampleBase(ENABLE_VALIDATION) {
		title = "Accurate Rendering of Liquid-Crystals and\nInhomogeneous Optically Anisotropic Media";
		settings.overlay = true;

		camera.type = Camera::CameraType::lookat;
		camera.setPerspective(60.0f, (float)width / (float)height, .5f, 512.f);
		camera.setRotation({ -30.0f, 90.0f, 0.0f });
		camera.setPosition({ .0f, 1.f, -3.5f });

		reload_profiles();
	}

	~VulkanExample()
	{
		// Clean up used Vulkan resources 
		vkDestroyPipeline(device, pipelines.scene, nullptr);
		vkDestroyPipeline(device, pipelines.composition, nullptr);
		vkDestroyPipeline(device, pipelines.object, nullptr);
		vkDestroyPipeline(device, pipelines.render, nullptr);
		vkDestroyPipeline(device, pipelines.volume, nullptr);
		vkDestroyPipelineLayout(device, pipelineLayouts.scene, nullptr);
		vkDestroyPipelineLayout(device, pipelineLayouts.composition, nullptr);
		vkDestroyDescriptorSetLayout(device, descriptorSetLayouts.scene, nullptr);
		vkDestroyDescriptorSetLayout(device, descriptorSetLayouts.composition, nullptr);

		models.object.destroy();
		models.cylinder.destroy();
		models.sphere.destroy();

		uniformBuffers.object.destroy();
		uniformBuffers.params.destroy();

		vkDestroySemaphore(device, offscreen.semaphore, nullptr);
		vkDestroyRenderPass(device, offscreen.renderPass, nullptr);
		vkDestroyFramebuffer(device, offscreen.frameBuffer, nullptr);
		vkDestroySampler(device, offscreen.sampler, nullptr);

		offscreen.depth.destroy(device);
		offscreen.color.destroy(device);
	}

	std::vector<double> readCSVline(std::istream& str) {
		std::vector<double> result;
		std::string l;
		std::getline(str,l);

		std::stringstream ss(l);
		std::string s;
		while (std::getline(ss, s, ','))
			result.push_back(std::stod(s));
		return result;

	}

	void resizeAssets()
	{}

	void loadAssets()
	{
		models.object.loadFromFile(getAssetPath() + "cube.obj", vertexLayout, 0.075f, vulkanDevice, queue);
		models.cylinder.loadFromFile(getAssetPath() + "cylinder.obj", vertexLayout, 0.075f, vulkanDevice, queue);
		models.sphere.loadFromFile(getAssetPath() + "sphere.obj", vertexLayout, 1.f / 19.73708f, vulkanDevice, queue);
	}

	void reBuildCommandBuffers()
	{
		if (!checkCommandBuffers())
		{
			destroyCommandBuffers();
			createCommandBuffers();
		}
		buildCommandBuffers();
	}

	void buildCommandBuffers()
	{		
		{
			if (offscreen.cmdBuffer == VK_NULL_HANDLE)
			{
				offscreen.cmdBuffer = VulkanExampleBase::createCommandBuffer(VK_COMMAND_BUFFER_LEVEL_PRIMARY, false);
			}
			if (offscreen.semaphore == VK_NULL_HANDLE)
			{
				VkSemaphoreCreateInfo semaphoreCreateInfo = vks::initializers::semaphoreCreateInfo();
				VK_CHECK_RESULT(vkCreateSemaphore(device, &semaphoreCreateInfo, nullptr, &offscreen.semaphore));
			}

			VkCommandBufferBeginInfo cmdBufInfo = vks::initializers::commandBufferBeginInfo();

			// Clear values for all attachments written in the fragment sahder
			std::array<VkClearValue, 2> clearValues;
			clearValues[0].color = defaultClearColor;
			clearValues[1].depthStencil = { 1.0f, 0 };

			VkRenderPassBeginInfo renderPassBeginInfo = vks::initializers::renderPassBeginInfo();
			renderPassBeginInfo.renderPass = offscreen.renderPass;
			renderPassBeginInfo.framebuffer = offscreen.frameBuffer;
			renderPassBeginInfo.renderArea.extent.width = offscreen.width;
			renderPassBeginInfo.renderArea.extent.height = offscreen.height;
			renderPassBeginInfo.clearValueCount = clearValues.size();
			renderPassBeginInfo.pClearValues = clearValues.data();

			VK_CHECK_RESULT(vkBeginCommandBuffer(offscreen.cmdBuffer, &cmdBufInfo));

			vkCmdBeginRenderPass(offscreen.cmdBuffer, &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

			VkViewport viewport = vks::initializers::viewport((float)offscreen.width, (float)offscreen.height, 0.0f, 1.0f);
			vkCmdSetViewport(offscreen.cmdBuffer, 0, 1, &viewport);
			VkRect2D scissor = vks::initializers::rect2D(offscreen.width, offscreen.height, 0, 0);
			vkCmdSetScissor(offscreen.cmdBuffer, 0, 1, &scissor);
			VkDeviceSize offsets[1] = { 0 };

			vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.scene, 0, 1, &descriptorSets.scene, 0, NULL);
			vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.scene);
			vkCmdBindVertexBuffers(offscreen.cmdBuffer, VERTEX_BUFFER_BIND_ID, 1, &models.cylinder.vertices.buffer, offsets);
			vkCmdBindIndexBuffer(offscreen.cmdBuffer, models.cylinder.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(offscreen.cmdBuffer, models.cylinder.indexCount, 1, 0, 0, 0);

			vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.scene, 0, 1, &descriptorSets.scene, 0, NULL);
			vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.render);
			vkCmdBindVertexBuffers(offscreen.cmdBuffer, 0, 1, &models.object.vertices.buffer, offsets);
			vkCmdBindIndexBuffer(offscreen.cmdBuffer, models.object.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(offscreen.cmdBuffer, models.object.indexCount, 1, 0, 0, 0);

			vkCmdNextSubpass(offscreen.cmdBuffer, VK_SUBPASS_CONTENTS_INLINE);

			if (render_props) {
				const auto& model = current_type == profile_type_t::slab || current_type == profile_type_t::heat ? models.object : models.sphere;

				vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.scene, 0, 1, &descriptorSets.scene, 0, NULL);
				vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.object);
				vkCmdBindVertexBuffers(offscreen.cmdBuffer, 0, 1, &model.vertices.buffer, offsets);
				vkCmdBindIndexBuffer(offscreen.cmdBuffer, model.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
				vkCmdDrawIndexed(offscreen.cmdBuffer, model.indexCount, 1, 0, 0, 0);

				vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.scene, 0, 1, &descriptorSets.scene, 0, NULL);
				vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.volume);
				vkCmdBindVertexBuffers(offscreen.cmdBuffer, 0, 1, &models.object.vertices.buffer, offsets);
				vkCmdBindIndexBuffer(offscreen.cmdBuffer, models.object.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
				vkCmdDrawIndexed(offscreen.cmdBuffer, models.object.indexCount, 1, 0, 0, 0);
			}

			vkCmdEndRenderPass(offscreen.cmdBuffer);

			VK_CHECK_RESULT(vkEndCommandBuffer(offscreen.cmdBuffer));
		}

		VkCommandBufferBeginInfo cmdBufInfo = vks::initializers::commandBufferBeginInfo();

		VkClearValue clearValues[2];
		clearValues[0].color = { { 0.0f, 0.0f, 0.0f, 0.0f } };
		clearValues[1].depthStencil = { 1.0f, 0 };

		VkRenderPassBeginInfo renderPassBeginInfo = vks::initializers::renderPassBeginInfo();
		renderPassBeginInfo.renderPass = renderPass;
		renderPassBeginInfo.renderArea.offset.x = 0;
		renderPassBeginInfo.renderArea.offset.y = 0;
		renderPassBeginInfo.renderArea.extent.width = width;
		renderPassBeginInfo.renderArea.extent.height = height;
		renderPassBeginInfo.clearValueCount = 2;
		renderPassBeginInfo.pClearValues = clearValues;

		for (int32_t i = 0; i < drawCmdBuffers.size(); ++i)
		{
			// Set target frame buffer
			renderPassBeginInfo.framebuffer = frameBuffers[i];

			VK_CHECK_RESULT(vkBeginCommandBuffer(drawCmdBuffers[i], &cmdBufInfo));

			const VkDeviceSize offsets[1] = { 0 };
			vkCmdBeginRenderPass(drawCmdBuffers[i], &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);
			VkViewport viewport = vks::initializers::viewport((float)width, (float)height, 0.0f, 1.0f);
			vkCmdSetViewport(drawCmdBuffers[i], 0, 1, &viewport);
			VkRect2D scissor = vks::initializers::rect2D(width, height,	0, 0);
			vkCmdSetScissor(drawCmdBuffers[i], 0, 1, &scissor);
			
			vkCmdBindDescriptorSets(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.composition, 0, 1, &descriptorSets.composition, 0, NULL);
			vkCmdBindPipeline(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.composition);
			vkCmdDraw(drawCmdBuffers[i], 3, 1, 0, 0);

			vkCmdEndRenderPass(drawCmdBuffers[i]);

			VK_CHECK_RESULT(vkEndCommandBuffer(drawCmdBuffers[i]));
		}
	}

	bool saveScreenshot(const char *filename) {
		const auto data = captureBackbuffer();
		if (!data)
			return false;

		auto p = std::make_unique<uint8_t[]>(width*height*3);
		for (uint32_t y = 0; y < height; y++) {
			for (uint32_t x = 0; x < width; x++) {
				const glm::vec3 c0 = data[y*width+x];
				glm::vec4 c1 = glm::vec4{ c0.x, c0.y, c0.z, .0f };
				auto packed = glm::packUnorm4x8(c1);
#if defined(_WIN32)
				const uint32_t order0 = 2, order1 = 1, order2 = 0;
#else
				const uint32_t order0 = 0, order1 = 1, order2 = 2;
#endif
				p[(y*width+x)*3+order0] = static_cast<uint8_t>((packed>>0) &0xFF);
				p[(y*width+x)*3+order1] = static_cast<uint8_t>((packed>>8) &0xFF);
				p[(y*width+x)*3+order2] = static_cast<uint8_t>((packed>>16)&0xFF);
			}
		}
		
		if (write_2d(filename, p.get(), 3, width, height))
			std::cout << "Screenshot saved to '" << filename << "'" << std::endl;

		return true;
	}

	std::unique_ptr<glm::vec3[]> captureBackbuffer() {
		// Source for the copy is the last rendered swapchain image
		VkImage srcImage = offscreen.color.image;
	
		// Create the linear tiled destination image to copy to and to read the memory from
		VkImageCreateInfo imageCreateCI(vks::initializers::imageCreateInfo());
		imageCreateCI.imageType = VK_IMAGE_TYPE_2D;
		// Note that vkCmdBlitImage (if supported) will also do format conversions if the swapchain color format would differ
		imageCreateCI.format = VK_FORMAT_R32G32B32A32_SFLOAT;
		imageCreateCI.extent.width = width;
		imageCreateCI.extent.height = height;
		imageCreateCI.extent.depth = 1;
		imageCreateCI.arrayLayers = 1;
		imageCreateCI.mipLevels = 1;
		imageCreateCI.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
		imageCreateCI.samples = VK_SAMPLE_COUNT_1_BIT;
		imageCreateCI.tiling = VK_IMAGE_TILING_LINEAR;
		imageCreateCI.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT;
		// Create the image
		VkImage dstImage;
		VK_CHECK_RESULT(vkCreateImage(device, &imageCreateCI, nullptr, &dstImage));
		// Create memory to back up the image
		VkMemoryRequirements memRequirements;
		VkMemoryAllocateInfo memAllocInfo(vks::initializers::memoryAllocateInfo());
		VkDeviceMemory dstImageMemory;
		vkGetImageMemoryRequirements(device, dstImage, &memRequirements);
		memAllocInfo.allocationSize = memRequirements.size;
		// Memory must be host visible to copy from
		memAllocInfo.memoryTypeIndex = vulkanDevice->getMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(device, &memAllocInfo, nullptr, &dstImageMemory));
		VK_CHECK_RESULT(vkBindImageMemory(device, dstImage, dstImageMemory, 0));

		// Do the actual blit from the swapchain image to our host visible destination image
		VkCommandBuffer copyCmd = vulkanDevice->createCommandBuffer(VK_COMMAND_BUFFER_LEVEL_PRIMARY, true);

		VkImageMemoryBarrier imageMemoryBarrier = vks::initializers::imageMemoryBarrier();
		
		// Transition destination image to transfer destination layout
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			dstImage,
			0,
			VK_ACCESS_TRANSFER_WRITE_BIT,
			VK_IMAGE_LAYOUT_UNDEFINED,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		// Transition swapchain image from present to transfer source layout
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			srcImage,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_ACCESS_TRANSFER_READ_BIT,
			VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
			VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		VkImageCopy imageCopyRegion{};
		imageCopyRegion.srcSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
		imageCopyRegion.srcSubresource.layerCount = 1;
		imageCopyRegion.dstSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
		imageCopyRegion.dstSubresource.layerCount = 1;
		imageCopyRegion.extent.width = width;
		imageCopyRegion.extent.height = height;
		imageCopyRegion.extent.depth = 1;

		// Issue the copy command
		vkCmdCopyImage(
			copyCmd,
			srcImage, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			dstImage, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			1,
			&imageCopyRegion);

		// Transition destination image to general layout, which is the required layout for mapping the image memory later on
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			dstImage,
			VK_ACCESS_TRANSFER_WRITE_BIT,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			VK_IMAGE_LAYOUT_GENERAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		// Transition back the swap chain image after the blit is done
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			srcImage,
			VK_ACCESS_TRANSFER_READ_BIT,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		vulkanDevice->flushCommandBuffer(copyCmd, queue);

		// Get layout of the image (including row pitch)
		VkImageSubresource subResource { VK_IMAGE_ASPECT_COLOR_BIT, 0, 0 };
		VkSubresourceLayout subResourceLayout;
		vkGetImageSubresourceLayout(device, dstImage, &subResource, &subResourceLayout);

		// Map image memory so we can start copying from it
		const char* data;
		vkMapMemory(device, dstImageMemory, 0, VK_WHOLE_SIZE, 0, (void**)&data);
		data += subResourceLayout.offset;

		auto p = std::make_unique<glm::vec3[]>(width*height);
		for (uint32_t y = 0; y < height; y++) {
			const auto *row = (const glm::vec4*)data;
			for (uint32_t x = 0; x < width; x++) {
				const glm::vec3 c0 = glm::vec3(1.f) - glm::exp(-glm::vec3{ *row } * uboParams.exposure);
				const glm::vec4 c1 = glm::vec4(glm::pow(c0, glm::vec3(1.f / 2.2f)), 1.f);
				p[y*width+x] = glm::vec3{ c1.x, c1.y, c1.z };
				++row;
			}
			data += subResourceLayout.rowPitch;
		}

		// Clean up resources
		vkUnmapMemory(device, dstImageMemory);
		vkFreeMemory(device, dstImageMemory, nullptr);
		vkDestroyImage(device, dstImage, nullptr);

		return p;
	}

	void createAttachment(VkFormat format, VkImageUsageFlagBits usage, FrameBufferAttachment *attachment) {
		VkImageAspectFlags aspectMask = 0;
		VkImageLayout imageLayout;

		attachment->format = format;

		if (usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT) {
			aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
			imageLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
		}
		if (usage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT) {
			aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
			imageLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
		}

		assert(aspectMask > 0);

		VkImageCreateInfo image = vks::initializers::imageCreateInfo();
		image.imageType = VK_IMAGE_TYPE_2D;
		image.format = format;
		image.extent.width = offscreen.width;
		image.extent.height = offscreen.height;
		image.extent.depth = 1;
		image.mipLevels = 1;
		image.arrayLayers = 1;
		image.samples = VK_SAMPLE_COUNT_1_BIT;
		image.tiling = VK_IMAGE_TILING_OPTIMAL;
		image.usage = usage | VK_IMAGE_USAGE_SAMPLED_BIT;

		VkMemoryAllocateInfo memAlloc = vks::initializers::memoryAllocateInfo();
		VkMemoryRequirements memReqs;

		VK_CHECK_RESULT(vkCreateImage(device, &image, nullptr, &attachment->image));
		vkGetImageMemoryRequirements(device, attachment->image, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		memAlloc.memoryTypeIndex = vulkanDevice->getMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(device, &memAlloc, nullptr, &attachment->mem));
		VK_CHECK_RESULT(vkBindImageMemory(device, attachment->image, attachment->mem, 0));

		VkImageViewCreateInfo imageView = vks::initializers::imageViewCreateInfo();
		imageView.viewType = VK_IMAGE_VIEW_TYPE_2D;
		imageView.format = format;
		imageView.subresourceRange = {};
		imageView.subresourceRange.aspectMask = aspectMask;
		imageView.subresourceRange.baseMipLevel = 0;
		imageView.subresourceRange.levelCount = 1;
		imageView.subresourceRange.baseArrayLayer = 0;
		imageView.subresourceRange.layerCount = 1;
		imageView.image = attachment->image;
		VK_CHECK_RESULT(vkCreateImageView(device, &imageView, nullptr, &attachment->view));
	}

	void setupFrameBuffer() override {
		VulkanExampleBase::setupFrameBuffer();

		offscreen.width = width;
		offscreen.height = height;

		// Two floating point color buffers
		createAttachment(VK_FORMAT_R32G32B32A32_SFLOAT, static_cast<VkImageUsageFlagBits>(VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT), &offscreen.color);
		// Depth attachment
		createAttachment(depthFormat, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, &offscreen.depth);

		// Set up separate renderpass with references to the color and depth attachments
		std::array<VkAttachmentDescription, 2> attachmentDescs = {};

		// Init attachment properties
		for (uint32_t i = 0; i < attachmentDescs.size(); ++i) {
			attachmentDescs[i].samples = VK_SAMPLE_COUNT_1_BIT;
			attachmentDescs[i].loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
			attachmentDescs[i].storeOp = VK_ATTACHMENT_STORE_OP_STORE;
			attachmentDescs[i].stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
			attachmentDescs[i].stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
			attachmentDescs[i].initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
		}
		attachmentDescs[0].finalLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		attachmentDescs[0].format = offscreen.color.format;
		attachmentDescs[1].finalLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
		attachmentDescs[1].format = offscreen.depth.format;
		attachmentDescs[1].loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachmentDescs[1].storeOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;

		std::vector<VkAttachmentReference> colorReferences;
		colorReferences.push_back({ 0, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL });

		VkAttachmentReference depthReference = {};
		depthReference.attachment = 1;
		depthReference.layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkSubpassDescription subpass[2] = {};
		subpass[0].pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpass[0].pColorAttachments = colorReferences.data();
		subpass[0].colorAttachmentCount = 1;
		subpass[0].pDepthStencilAttachment = &depthReference;
		subpass[1].pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpass[1].pColorAttachments = colorReferences.data();
		subpass[1].colorAttachmentCount = colorReferences.size();
		subpass[1].pDepthStencilAttachment = &depthReference;

		// Use subpass dependencies for attachment layput transitions
		std::array<VkSubpassDependency, 3> dependencies;

		dependencies[0].srcSubpass = VK_SUBPASS_EXTERNAL;
		dependencies[0].dstSubpass = 0;
		dependencies[0].srcStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[0].dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[0].srcAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[0].dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[0].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[1].srcSubpass = 0;
		dependencies[1].dstSubpass = 1;
		dependencies[1].srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[1].dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[1].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[2].srcSubpass = 1;
		dependencies[2].dstSubpass = VK_SUBPASS_EXTERNAL;
		dependencies[2].srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[2].dstStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[2].srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[2].dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[2].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		VkRenderPassCreateInfo renderPassInfo = {};
		renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
		renderPassInfo.pAttachments = attachmentDescs.data();
		renderPassInfo.attachmentCount = static_cast<uint32_t>(attachmentDescs.size());
		renderPassInfo.subpassCount = 2;
		renderPassInfo.pSubpasses = subpass;
		renderPassInfo.dependencyCount = dependencies.size();
		renderPassInfo.pDependencies = dependencies.data();

		VK_CHECK_RESULT(vkCreateRenderPass(device, &renderPassInfo, nullptr, &offscreen.renderPass));

		std::array<VkImageView, 2> attachments;
		attachments[0] = offscreen.color.view;
		attachments[1] = offscreen.depth.view;

		VkFramebufferCreateInfo fbufCreateInfo = {};
		fbufCreateInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
		fbufCreateInfo.pNext = NULL;
		fbufCreateInfo.renderPass = offscreen.renderPass;
		fbufCreateInfo.pAttachments = attachments.data();
		fbufCreateInfo.attachmentCount = static_cast<uint32_t>(attachments.size());
		fbufCreateInfo.width = offscreen.width;
		fbufCreateInfo.height = offscreen.height;
		fbufCreateInfo.layers = 1;
		VK_CHECK_RESULT(vkCreateFramebuffer(device, &fbufCreateInfo, nullptr, &offscreen.frameBuffer));

		// Create sampler to sample from the color attachments
		VkSamplerCreateInfo sampler = vks::initializers::samplerCreateInfo();
		sampler.magFilter = VK_FILTER_NEAREST;
		sampler.minFilter = VK_FILTER_NEAREST;
		sampler.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
		sampler.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		sampler.addressModeV = sampler.addressModeU;
		sampler.addressModeW = sampler.addressModeU;
		sampler.mipLodBias = 0.0f;
		sampler.maxAnisotropy = 1.0f;
		sampler.minLod = 0.0f;
		sampler.maxLod = 1.0f;
		sampler.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;
		VK_CHECK_RESULT(vkCreateSampler(device, &sampler, nullptr, &offscreen.sampler));

		resizeAssets();
	}

	void prepareVertices() {
		// Binding description
		vertices.bindingDescriptions.resize(1);
		vertices.bindingDescriptions[0] =
			vks::initializers::vertexInputBindingDescription(
				VERTEX_BUFFER_BIND_ID,
				vertexLayout.stride(),
				VK_VERTEX_INPUT_RATE_VERTEX);

		// Attribute descriptions
		// Describes memory layout and shader positions
		vertices.attributeDescriptions.resize(4);
		// Location 0 : Position
		vertices.attributeDescriptions[0] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				0,
				VK_FORMAT_R32G32B32_SFLOAT,
				0);
		// Location 1 : Normal
		vertices.attributeDescriptions[1] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				1,
				VK_FORMAT_R32G32B32_SFLOAT,
				sizeof(float) * 3);
		// Location 2 : Texture coordinates
		vertices.attributeDescriptions[2] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				2,
				VK_FORMAT_R32G32_SFLOAT,
				sizeof(float) * 6);
		// Location 3 : Color
		vertices.attributeDescriptions[3] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				3,
				VK_FORMAT_R32G32B32_SFLOAT,
				sizeof(float) * 8);

		vertices.inputState = vks::initializers::pipelineVertexInputStateCreateInfo();
		vertices.inputState.vertexBindingDescriptionCount = vertices.bindingDescriptions.size();
		vertices.inputState.pVertexBindingDescriptions = vertices.bindingDescriptions.data();
		vertices.inputState.vertexAttributeDescriptionCount = vertices.attributeDescriptions.size();
		vertices.inputState.pVertexAttributeDescriptions = vertices.attributeDescriptions.data();
	}

	void setupDescriptorPool() {
		// Example uses one ubo and one image sampler
		std::vector<VkDescriptorPoolSize> poolSizes = {
			vks::initializers::descriptorPoolSize(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 8),
			vks::initializers::descriptorPoolSize(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 12)
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo =
			vks::initializers::descriptorPoolCreateInfo(
				poolSizes.size(),
				poolSizes.data(),
				3);

		VK_CHECK_RESULT(vkCreateDescriptorPool(device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}

	void setupDescriptorSetLayout() {
		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings = {
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0),
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 1),
		};

		VkDescriptorSetLayoutCreateInfo descriptorLayout =
			vks::initializers::descriptorSetLayoutCreateInfo(
				setLayoutBindings.data(),
				setLayoutBindings.size());

		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(device, &descriptorLayout, nullptr, &descriptorSetLayouts.scene));

		VkPipelineLayoutCreateInfo pPipelineLayoutCreateInfo =
			vks::initializers::pipelineLayoutCreateInfo(
				&descriptorSetLayouts.scene,
				1);
		pPipelineLayoutCreateInfo.pushConstantRangeCount = 0;

		VK_CHECK_RESULT(vkCreatePipelineLayout(device, &pPipelineLayoutCreateInfo, nullptr, &pipelineLayouts.scene));

		{
			setLayoutBindings = {
				vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT, 0),
				vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_FRAGMENT_BIT, 1)
			};

			descriptorLayout = vks::initializers::descriptorSetLayoutCreateInfo(setLayoutBindings.data(), static_cast<uint32_t>(setLayoutBindings.size()));
			VK_CHECK_RESULT(vkCreateDescriptorSetLayout(device, &descriptorLayout, nullptr, &descriptorSetLayouts.composition));

			pPipelineLayoutCreateInfo = vks::initializers::pipelineLayoutCreateInfo(&descriptorSetLayouts.composition, 1);
			pPipelineLayoutCreateInfo.pushConstantRangeCount = 0;
			VK_CHECK_RESULT(vkCreatePipelineLayout(device, &pPipelineLayoutCreateInfo, nullptr, &pipelineLayouts.composition));
		}
	}

	void setupDescriptorSet() {
		VkDescriptorSetAllocateInfo allocInfo =
			vks::initializers::descriptorSetAllocateInfo(
				descriptorPool,
				&descriptorSetLayouts.scene,
				1);

		VK_CHECK_RESULT(vkAllocateDescriptorSets(device, &allocInfo, &descriptorSets.scene));

		std::vector<VkWriteDescriptorSet> writeDescriptorSets = {
			vks::initializers::writeDescriptorSet(descriptorSets.scene, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 0, &uniformBuffers.object.descriptor),
			vks::initializers::writeDescriptorSet(descriptorSets.scene, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1, &uniformBuffers.params.descriptor),
		};

		vkUpdateDescriptorSets(device, writeDescriptorSets.size(), writeDescriptorSets.data(), 0, NULL);

		{
			allocInfo =	vks::initializers::descriptorSetAllocateInfo(descriptorPool, &descriptorSetLayouts.composition, 1);
			VK_CHECK_RESULT(vkAllocateDescriptorSets(device, &allocInfo, &descriptorSets.composition));

			std::vector<VkDescriptorImageInfo> colorDescriptors = {
				vks::initializers::descriptorImageInfo(offscreen.sampler, offscreen.color.view, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL),
			};

			writeDescriptorSets = {
				vks::initializers::writeDescriptorSet(descriptorSets.composition, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 0, &colorDescriptors[0]),
				vks::initializers::writeDescriptorSet(descriptorSets.composition, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1, &uniformBuffers.params.descriptor)
			};
			vkUpdateDescriptorSets(device, static_cast<uint32_t>(writeDescriptorSets.size()), writeDescriptorSets.data(), 0, NULL);
		}
	}

	void preparePipelines() {
		const auto profile_type = static_cast<std::underlying_type_t<decltype(current_type)>>(current_type);
		float droplet_scale = 0.14f;
		if (current_type == profile_type_t::droplet || current_type == profile_type_t::shell)
			droplet_scale *= 18 * 3.8;

		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState =
			vks::initializers::pipelineInputAssemblyStateCreateInfo(
				VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST,
				0,
				VK_FALSE);

		VkPipelineRasterizationStateCreateInfo rasterizationState =
			vks::initializers::pipelineRasterizationStateCreateInfo(
				VK_POLYGON_MODE_FILL,
				VK_CULL_MODE_BACK_BIT,
				VK_FRONT_FACE_CLOCKWISE,
				0);

		VkPipelineColorBlendAttachmentState blendAttachmentState[1] = {
			vks::initializers::pipelineColorBlendAttachmentState(0xf, VK_FALSE),
		};

		VkPipelineColorBlendStateCreateInfo colorBlendState =
			vks::initializers::pipelineColorBlendStateCreateInfo(
				1,
				blendAttachmentState);

		VkPipelineDepthStencilStateCreateInfo depthStencilState =
			vks::initializers::pipelineDepthStencilStateCreateInfo(
				VK_TRUE,
				VK_TRUE,
				VK_COMPARE_OP_LESS_OR_EQUAL);

		VkPipelineViewportStateCreateInfo viewportState =
			vks::initializers::pipelineViewportStateCreateInfo(1, 1, 0);

		VkPipelineMultisampleStateCreateInfo multisampleState =
			vks::initializers::pipelineMultisampleStateCreateInfo(
				VK_SAMPLE_COUNT_1_BIT,
				0);

		std::vector<VkDynamicState> dynamicStateEnables = {
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};
		VkPipelineDynamicStateCreateInfo dynamicState =
			vks::initializers::pipelineDynamicStateCreateInfo(
				dynamicStateEnables.data(),
				dynamicStateEnables.size(),
				0);

		// Load shaders
		VkGraphicsPipelineCreateInfo pipelineCreateInfo =
			vks::initializers::pipelineCreateInfo(
				pipelineLayouts.scene,
				offscreen.renderPass,
				0);
		pipelineCreateInfo.pVertexInputState = &vertices.inputState;
		pipelineCreateInfo.pInputAssemblyState = &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState = &rasterizationState;
		pipelineCreateInfo.pColorBlendState = &colorBlendState;
		pipelineCreateInfo.pMultisampleState = &multisampleState;
		pipelineCreateInfo.pViewportState = &viewportState;
		pipelineCreateInfo.pDepthStencilState = &depthStencilState;
		pipelineCreateInfo.pDynamicState = &dynamicState;
		pipelineCreateInfo.subpass = 0;

		{
			VkPipelineRasterizationStateCreateInfo rasterizationState =
				vks::initializers::pipelineRasterizationStateCreateInfo(
					VK_POLYGON_MODE_FILL,
					VK_CULL_MODE_FRONT_BIT,
					VK_FRONT_FACE_CLOCKWISE,
					0);

			constexpr int specializationConstants = 1;
			std::array<VkSpecializationMapEntry, specializationConstants> specializationMapEntries;
			specializationMapEntries[0].constantID = 0;
			specializationMapEntries[0].size = 4;
			specializationMapEntries[0].offset = 0;
			VkSpecializationInfo specializationInfo{};
			specializationInfo.dataSize = 4;
			specializationInfo.mapEntryCount = static_cast<uint32_t>(specializationMapEntries.size());
			specializationInfo.pMapEntries = specializationMapEntries.data();
			specializationInfo.pData = &white_light;

			std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/render/scene.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() + "../shaders/render/scene.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			shaderStages[1].pSpecializationInfo = &specializationInfo;

			pipelineCreateInfo.stageCount = shaderStages.size();
			pipelineCreateInfo.pStages = shaderStages.data();
			pipelineCreateInfo.pRasterizationState = &rasterizationState;
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.scene));
		}
		pipelineCreateInfo.pRasterizationState = &rasterizationState;
		{
			constexpr int specializationConstants = 4;
			std::array<VkSpecializationMapEntry, specializationConstants> specializationMapEntries;
			specializationMapEntries[0].constantID = 0;
			specializationMapEntries[0].size = 4;
			specializationMapEntries[0].offset = 0;
			specializationMapEntries[1].constantID = 1;
			specializationMapEntries[1].size = 4;
			specializationMapEntries[1].offset = 4;
			specializationMapEntries[2].constantID = 2;
			specializationMapEntries[2].size = 4;
			specializationMapEntries[2].offset = 8;
			specializationMapEntries[3].constantID = 3;
			specializationMapEntries[3].size = 4;
			specializationMapEntries[3].offset = 12;
			int32_t constants[specializationConstants] = {};
			constants[0] = view_mode;
			constants[1] = layers;
			constants[2] = white_light;
			constants[3] = profile_type;
			VkSpecializationInfo specializationInfo{};
			specializationInfo.dataSize = sizeof(constants);
			specializationInfo.mapEntryCount = static_cast<uint32_t>(specializationMapEntries.size());
			specializationInfo.pMapEntries = specializationMapEntries.data();
			specializationInfo.pData = constants;

			std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/render/sem.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			if (current_material.profile)
				shaderStages[1] = loadShader(getAssetPath() + spirv_dst.string(), VK_SHADER_STAGE_FRAGMENT_BIT);
			else
				shaderStages[1] = loadShader(getAssetPath() + spirv_src[current_version].string()+".spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			shaderStages[1].pSpecializationInfo = &specializationInfo;

			pipelineCreateInfo.stageCount = shaderStages.size();
			pipelineCreateInfo.pStages = shaderStages.data();
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.render));
		}
		{
			constexpr int specializationConstants = 7;
			std::array<VkSpecializationMapEntry, specializationConstants> specializationMapEntries;
			for (int i=0;i<specializationConstants;++i) {
				specializationMapEntries[i].constantID = i;
				specializationMapEntries[i].size = 4;
				specializationMapEntries[i].offset = 4*i;
			}
			glm::vec3 constants[3] = {};
			if (current_type == profile_type_t::slab || current_type == profile_type_t::heat) {
				constants[0] = plate_scale;
				constants[1] = plate_offset;
				*reinterpret_cast<uint32_t*>(reinterpret_cast<uint8_t*>(constants)+specializationMapEntries[6].offset) = 1;		// scale y with tau
			}
			else {
				constants[0] = glm::vec3{ droplet_scale };
				constants[1] = plate_offset - glm::vec3{0,1,0};
				*reinterpret_cast<uint32_t*>(reinterpret_cast<uint8_t*>(constants)+specializationMapEntries[6].offset) = 2;		// scale uniformly with diameter
			}
			VkSpecializationInfo specializationInfo{};
			specializationInfo.dataSize = specializationConstants * 4;
			specializationInfo.mapEntryCount = static_cast<uint32_t>(specializationMapEntries.size());
			specializationInfo.pMapEntries = specializationMapEntries.data();
			specializationInfo.pData = constants;

			std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/render/object.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() + "../shaders/render/object.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			shaderStages[0].pSpecializationInfo = &specializationInfo;
			shaderStages[1].pSpecializationInfo = &specializationInfo;

			VkPipelineDepthStencilStateCreateInfo depthStencilState = vks::initializers::pipelineDepthStencilStateCreateInfo(VK_TRUE, VK_FALSE, VK_COMPARE_OP_LESS_OR_EQUAL);

			VkPipelineColorBlendAttachmentState blendAttachmentState = {};
			blendAttachmentState.blendEnable = VK_TRUE;
			blendAttachmentState.srcColorBlendFactor = VK_BLEND_FACTOR_SRC_ALPHA;
			blendAttachmentState.dstColorBlendFactor = VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
			blendAttachmentState.colorBlendOp = VK_BLEND_OP_ADD;
			blendAttachmentState.srcAlphaBlendFactor = VK_BLEND_FACTOR_ONE;
			blendAttachmentState.dstAlphaBlendFactor = VK_BLEND_FACTOR_ZERO;
			blendAttachmentState.alphaBlendOp = VK_BLEND_OP_ADD;
			blendAttachmentState.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
			VkPipelineColorBlendStateCreateInfo colorBlendState = vks::initializers::pipelineColorBlendStateCreateInfo(1, &blendAttachmentState);

			pipelineCreateInfo.pColorBlendState = &colorBlendState;
			pipelineCreateInfo.pDepthStencilState = &depthStencilState;
			pipelineCreateInfo.stageCount = shaderStages.size();
			pipelineCreateInfo.pStages = shaderStages.data();
			pipelineCreateInfo.subpass = 1;
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.object));
		}
		{
			constexpr int specializationConstants = 7;
			std::array<VkSpecializationMapEntry, specializationConstants> specializationMapEntries;
			for (int i=0;i<specializationConstants;++i) {
				specializationMapEntries[i].constantID = i;
				specializationMapEntries[i].size = 4;
				specializationMapEntries[i].offset = 4*i;
			}
			glm::vec3 constants[3] = {};
			constants[0] = volume_scale;
			constants[1] = volume_offset;
			*reinterpret_cast<uint32_t*>(reinterpret_cast<uint8_t*>(constants)+specializationMapEntries[6].offset) = 0;		// do not scale with tau
			VkSpecializationInfo specializationInfo{};
			specializationInfo.dataSize = specializationConstants*4;
			specializationInfo.mapEntryCount = static_cast<uint32_t>(specializationMapEntries.size());
			specializationInfo.pMapEntries = specializationMapEntries.data();
			specializationInfo.pData = constants;

			std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/render/object.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() + "../shaders/render/volume.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			shaderStages[0].pSpecializationInfo = &specializationInfo;

			VkPipelineDepthStencilStateCreateInfo depthStencilState = vks::initializers::pipelineDepthStencilStateCreateInfo(VK_TRUE, VK_FALSE, VK_COMPARE_OP_LESS_OR_EQUAL);

			VkPipelineColorBlendAttachmentState blendAttachmentState = {};
			blendAttachmentState.blendEnable = VK_TRUE;
			blendAttachmentState.srcColorBlendFactor = VK_BLEND_FACTOR_SRC_ALPHA;
			blendAttachmentState.dstColorBlendFactor = VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
			blendAttachmentState.colorBlendOp = VK_BLEND_OP_ADD;
			blendAttachmentState.srcAlphaBlendFactor = VK_BLEND_FACTOR_ONE;
			blendAttachmentState.dstAlphaBlendFactor = VK_BLEND_FACTOR_ZERO;
			blendAttachmentState.alphaBlendOp = VK_BLEND_OP_ADD;
			blendAttachmentState.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
			VkPipelineColorBlendStateCreateInfo colorBlendState = vks::initializers::pipelineColorBlendStateCreateInfo(1, &blendAttachmentState);

			pipelineCreateInfo.pColorBlendState = &colorBlendState;
			pipelineCreateInfo.pDepthStencilState = &depthStencilState;
			pipelineCreateInfo.stageCount = shaderStages.size();
			pipelineCreateInfo.pStages = shaderStages.data();
			pipelineCreateInfo.subpass = 1;
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.volume));
		}
		pipelineCreateInfo.pColorBlendState = &colorBlendState;
		pipelineCreateInfo.pDepthStencilState = &depthStencilState;

		{
			VkPipelineVertexInputStateCreateInfo emptyInputState = vks::initializers::pipelineVertexInputStateCreateInfo();
			pipelineCreateInfo.pVertexInputState = &emptyInputState;

			VkPipelineRasterizationStateCreateInfo rasterizationState =
				vks::initializers::pipelineRasterizationStateCreateInfo(VK_POLYGON_MODE_FILL, VK_CULL_MODE_FRONT_BIT, VK_FRONT_FACE_COUNTER_CLOCKWISE);
			pipelineCreateInfo.pRasterizationState = &rasterizationState;

			// Final fullscreen composition pass pipeline
			std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/hdr/composition.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() + "../shaders/hdr/composition.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);

			pipelineCreateInfo.layout = pipelineLayouts.composition;
			pipelineCreateInfo.renderPass = renderPass;
			pipelineCreateInfo.stageCount = shaderStages.size();
			pipelineCreateInfo.pStages = shaderStages.data();
			pipelineCreateInfo.subpass = 0;
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.composition));
		}
	}

	void prepareUniformBuffers() {
		// Vertex shader uniform buffer block
		VK_CHECK_RESULT(vulkanDevice->createBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&uniformBuffers.object,
			sizeof(uboVS)));

		// Tonemapping params
		VK_CHECK_RESULT(vulkanDevice->createBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&uniformBuffers.params,
			sizeof(uboParams)));

		// Map persistent
		VK_CHECK_RESULT(uniformBuffers.object.map());
		VK_CHECK_RESULT(uniformBuffers.params.map());

		updateUniformBuffers();
	}

	void updateUniformBuffers() {
		uboVS.projection = camera.matrices.perspective;
		uboVS.modelView = camera.matrices.view;
		uboVS.invModelView = glm::inverseTranspose(uboVS.modelView);
		uboVS.cameraPosition = glm::vec4{ camera.position, 1 };
		uboVS.white_light = white_light;
		uboVS.profile_type = static_cast<std::underlying_type_t<decltype(current_type)>>(current_type);
		memcpy(uniformBuffers.object.mapped, &uboVS, sizeof(uboVS));
	}
	void updateParams() {
		uboParams.time = static_cast<float>((std::chrono::steady_clock::now() - start_time).count()) * std::chrono::steady_clock::period::num / std::chrono::steady_clock::period::den;
		memcpy(uniformBuffers.params.mapped, &uboParams, sizeof(uboParams));
	}

	void draw() {
		VulkanExampleBase::prepareFrame();

		submitInfo.pWaitSemaphores = &semaphores.presentComplete;
		submitInfo.pSignalSemaphores = &offscreen.semaphore;
		submitInfo.commandBufferCount = 1;
		submitInfo.pCommandBuffers = &offscreen.cmdBuffer;
		VK_CHECK_RESULT(vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE));

		submitInfo.pWaitSemaphores = &offscreen.semaphore;
		submitInfo.pSignalSemaphores = &semaphores.renderComplete;
		submitInfo.commandBufferCount = 1;
		submitInfo.pCommandBuffers = &drawCmdBuffers[currentBuffer];
		VK_CHECK_RESULT(vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE));

		VulkanExampleBase::submitFrame();
	}

	void prepare() {
		VulkanExampleBase::prepare();

		loadAssets();
		
		prepareVertices();
		prepareUniformBuffers();
		setupDescriptorSetLayout();
		preparePipelines();
		setupDescriptorPool();
		setupDescriptorSet();
		buildCommandBuffers();

		updateParams();
		
		prepared = true;
	}

	virtual void render() override {
		if (!prepared)
			return;

		if (animate_wavelength) {
			const auto now = std::chrono::steady_clock::now();
			animate_wavelength_key += static_cast<float>((now - animate_wavelength_timestamp).count()) * .075f * std::chrono::steady_clock::period::num / std::chrono::steady_clock::period::den;
			if (animate_wavelength_key >= 1.f)
				animate_wavelength_key -= 2.f;
			uboVS.wavelength = glm::mix(380.f,750.f,glm::smoothstep(.0f,1.f,glm::abs(animate_wavelength_key)));
			animate_wavelength_timestamp = now;
			updateUniformBuffers();
		}
		
		if (screenshot_context && screenshot_context.value().update_camera) {
			// Place camera above the pattern
			camera.type = Camera::CameraType::firstperson;
			camera.setRotation({ -90.f, 90.f, .0f });
			camera.setPosition({ .0f, 2.92f, .0f });
			updateUniformBuffers();
		}

		updateParams();
		draw();

		if (screenshot_context) {
			saveScreenshot((std::string(getAssetPath()) + "../" + screenshot_context.value().name + ".png").c_str());

			view_mode = screenshot_context.value().prev_view_mode;
			white_light = screenshot_context.value().prev_white_light;
			render_props = screenshot_context.value().prev_render_props;
			if (screenshot_context.value().update_camera) {
				camera.type = Camera::CameraType::lookat;
				camera.setRotation(screenshot_context.value().prev_camera_rot);
				camera.setPosition(screenshot_context.value().prev_camera_pos);
				viewChanged();
			}

			vkDeviceWaitIdle(device);
			preparePipelines();
			reBuildCommandBuffers();

			screenshot_context = std::nullopt;
		}
	}

	virtual void viewChanged() override {
		if (-55.f > camera.rotation.x || camera.rotation.x > -10.f) {
			camera.rotation.x = glm::clamp(camera.rotation.x, -55.f, -10.f);
			camera.updateViewMatrix();
		}
		updateUniformBuffers();
	}
	virtual void windowResized() override {
		setupDescriptorSetLayout();
		preparePipelines();
		setupDescriptorPool();
		setupDescriptorSet();
		buildCommandBuffers();
		
		updateUniformBuffers();
	}

	virtual void OnUpdateUIOverlay(vks::UIOverlay *overlay) override {
		bool should_rebuild_pipelines = false;
		if (overlay->header("Settings")) {
			overlay->inputFloat("Exposure", &uboParams.exposure, 0.025f, 3);
			ImGui::SameLine();
			if (overlay->checkBox("Render props", &render_props))
				should_rebuild_pipelines = true;

			static int light_mode = 0;
			if (overlay->comboBox("Light", &light_mode, { std::string("Monochrome"), std::string("Animated Monochrome"), std::string("White E") })) {
				animate_wavelength = light_mode == 1;
				if (animate_wavelength) {
					animate_wavelength_timestamp = std::chrono::steady_clock::now();
					animate_wavelength_key = (animate_wavelength_key > .0f ? 1.f : -1.f) * inverse_smoothstep((uboVS.wavelength - 380.f) / (750.f - 380.f));
				}

				if (light_mode == 2 || white_light)
					should_rebuild_pipelines = true;
				white_light = light_mode == 2;

				updateUniformBuffers();
			}
			if (light_mode != 2) {
				ImGui::SameLine();
				if (overlay->sliderFloat("Wavelength (nm)", &uboVS.wavelength, 380.f, 750.f)) {
					animate_wavelength = false;
					light_mode = 0;

					updateUniformBuffers();
				}
			}

			overlay->sliderFloat("Incident polarization - 0: s 1: p", &uboParams.polarization, .0f, 1.f);
			if (overlay->comboBox("Filter", &view_mode, { std::string("No filter"), std::string("s-polarization filter"), std::string("p-polarization filter") }))
				should_rebuild_pipelines = true;
			ImGui::SameLine();
			static int render_mode = 0;
			if (overlay->comboBox("Render mode", &render_mode, { std::string("Normal"), 
																 std::string("Layered (1024 layers)"),
																 std::string("Layered (4096 layers)") })) {
				layers = render_mode == 0 ? 0 :
						 render_mode == 1 ? 1024 : 4096;
				should_rebuild_pipelines = true;
			}
		}

		const auto clamp_d = [](float d, profile_type_t t) {
			if (t == profile_type_t::slab || t == profile_type_t::heat)
				return glm::clamp(d, .1f, .8f);
			if (t == profile_type_t::droplet || t == profile_type_t::shell)
				return glm::clamp(d, .05f, .25f);
			return glm::clamp(d, 3.f, 15.f);
		};

		// Material manager
		if (overlay->header("Material")) {
			if (profiles.profiles.size()) {
				static int current_material_type = 0;
				static std::vector<std::string> current_material_types = {};

				profile_names[0] = "Current (" + (selected_material.profile ? selected_material.profile.name : "none") + ")";
				int32_t p = 0;
				ImGui::PushItemWidth(185.f);
				if (overlay->comboBox("Profile", &p, profile_names) && p>0) {
					selected_material.profile = profiles.profiles[p-1];
				}
				profile_names[0] = "";
				ImGui::PopItemWidth();

				if (selected_material.profile) {
					ImGui::SameLine();
					if (overlay->button("Update")) {
						if (compile_material(selected_material)) {
							current_material = selected_material;
							current_material_type = 0;
							current_type = current_material.profile.types[current_material_type];

							uboParams.d = clamp_d(uboParams.d, current_type);
							if (current_type == profile_type_t::heat)
								uboParams.c1 = glm::max(1.f, uboParams.c1);

							current_material_types.clear();
							for (const auto &type : selected_material.profile.types)
								current_material_types.emplace_back(profiles_manager::profile_type_name(type));

							should_rebuild_pipelines = true;
						}
						else
							std::cerr << "Material compilation failed!" << std::endl;
					}
				}
				if (current_material.profile) {
					ImGui::SameLine();
					if (overlay->button("Reset")) {
						selected_material = {};
						current_material = {};
						should_rebuild_pipelines = true;
					}
				}

				if (current_material.profile) {
					if (current_material_types.size()) {
						// Profile type selector
						if (overlay->comboBox("", &current_material_type, current_material_types)) {
							current_type = current_material.profile.types[current_material_type];
							uboParams.d = clamp_d(uboParams.d, current_type);
							if (current_type == profile_type_t::shell)
								uboParams.c2 = 5;

							should_rebuild_pipelines = true;
						}
						ImGui::SameLine();
					}
				}
			}

			if (current_type == profile_type_t::slab || current_type == profile_type_t::heat) {
				overlay->sliderFloat("\xCF\x84 (mm)", &uboParams.d, .0f, .8f);
				overlay->sliderFloat("Rotation", &uboParams.c2, -1, 1);
				if (current_type == profile_type_t::heat) {
					ImGui::SameLine();
					overlay->sliderFloat("Diffusion (pow)", &uboParams.c1, 1.f, 6.f);
				}
			}
			else {
				// Droplet
				if (current_type != profile_type_t::droplet && current_type != profile_type_t::shell) {
					overlay->sliderFloat("d (\xC2\xB5m)", &uboParams.d, .0f, 15.f);
					if (current_type == profile_type_t::droplet_axial)
						overlay->sliderFloat("Disclination ring", &uboParams.c2, .0f, 1.f);
				}
				else {
					overlay->sliderFloat("d (mm)", &uboParams.d, .0f, .25f);
					if (current_type == profile_type_t::shell)
						overlay->sliderFloat("Shell thickness (%)", &uboParams.c2, .0f, 20.f);
				}
			}

			if (current_type == profile_type_t::slab) {
				ImGui::TextColored(ImVec4(0.4, 0.8, 1, 1), "Control constants (if used)");
				overlay->sliderFloat("c0", &uboParams.c0, -1, 1);
				ImGui::SameLine();
				overlay->sliderFloat("c1", &uboParams.c1, -1, 1);
			}
		}

		// Profile manager (creating new profiles requires MATLAB)
		if (overlay->header("Profiles")) {
			ImGui::BeginChild("Profiles", {420, 130});
			bool shouldReloadProfiles = false;
			int i=0;
			for (const auto &p : profiles.profiles) {
				if (overlay->button((std::string("X##DeleteProfile")+std::to_string(++i)).c_str())) { 
					profiles.delete_profile(p);
					shouldReloadProfiles=true;
				}
				ImGui::SameLine();
				overlay->text(p.name.c_str());
			}
			ImGui::EndChild();
			if (shouldReloadProfiles)
				reload_profiles();
		}
		if (overlay->header("Create New Profile")) {
			static char name[100] = "test";
			overlay->text("Profile Name ");   ImGui::SameLine(140.f); ImGui::PushItemWidth(270.f); ImGui::InputText("##ProfileName", name, 100, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			
#ifdef _Matlab_Engine
			ImGui::BeginGroup();
			static char a[255] = "0.8*cos(h)*sin(c4)", c[255] = "0.8*cos(h)*cos(c4)", b[255] = "sqrt(1-a^2-c^2)", eo[255] = "1.5", ee[255] = "1.51-h*0.4", so[255] = "0", se[255] = "0";
			static int expansion_order = 3;
			static bool should_optimize=true;
			overlay->text(u8"α(h)=");   ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profilea",    a,    255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"β(h)=");   ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profileb",    b,    255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"γ(h)=");   ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profilec",    c,    255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"η_o=");    ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profileeo",   eo,   255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"η_e(h)="); ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profileee",   ee,   255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"σ_o");     ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profileso",   so,   255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			overlay->text(u8"σ_e(h)="); ImGui::SameLine(70.f);ImGui::PushItemWidth(330.f); ImGui::InputText("##Profilese",   se,   255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);	ImGui::PopItemWidth();
			ImGui::SliderInt("Order", &expansion_order, 2, 6); 
#ifdef _linux
			overlay->checkBox("Optimize", &should_optimize); ImGui::SameLine();
			overlay->text("(see readme)");
#else
			overlay->text("Optimization only available on Linux");
#endif

			if (overlay->button("Generate new profile"))
				profiles.start_new_profile_job(static_cast<std::uint32_t>(expansion_order),
											   should_optimize,
											   name, a, b, c, eo, ee, so, se);
			ImGui::EndGroup();
#endif

			ImGui::BeginGroup();
			static char import_path[255] = {};
			ImGui::PushItemWidth(320.f);  ImGui::InputTextWithHint("##PathImport", "Profile path", import_path, 255, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr); ImGui::PopItemWidth(); 
			if (overlay->button("Import MATLAB generated profile")) 
				profiles.start_new_import_job(name, import_path);
			ImGui::EndGroup();
		
			ImGui::Separator();
			ImGui::TextColored(ImVec4(0.4, 0.8, 1, 1), "Jobs");
			ImGui::BeginChild("Jobs", {420, 50});
			for (auto it = profiles.profile_jobs.begin() ; it!=profiles.profile_jobs.end();) {
				auto &p = *it;
				
				if (overlay->button((std::string("X##DeleteJob")+std::to_string(std::distance(profiles.profile_jobs.begin(),it))).c_str())) {
					profiles.stop_job(p);
					it = profiles.profile_jobs.erase(it);
					continue;
				}
				ImGui::SameLine(20.f);
				ImGui::PushItemWidth(140.f); overlay->text(p.profile.name.c_str()); ImGui::PopItemWidth();
				ImGui::SameLine(120.f);
				ImGui::PushItemWidth(90.f); 
				ImGui::TextColored(p.status == profiles_manager::profile_job_t::status_t::error      ? ImVec4(1,0.2,0.2,1)  :
								   p.status == profiles_manager::profile_job_t::status_t::successful ? ImVec4(0.5,1,0.65,1) : ImVec4(.8,.8,1.,1), p.status_text.c_str());
				ImGui::PopItemWidth(); 

				++it;
			}
			ImGui::EndChild();

			profiles.update_jobs();
		}

		if (overlay->header("Screenshot")) {
			static bool capture_pattern = false, force_white_light = false;
			static char screenshot_name[250] = {};
			ImGui::PushItemWidth(300.f);  ImGui::InputTextWithHint("##NameScreenshot", "Name", screenshot_name, 250, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr); ImGui::PopItemWidth(); 
			overlay->checkBox("Capture pattern", &capture_pattern);
			ImGui::SameLine();
			overlay->checkBox("White light", &force_white_light);

			if (overlay->button("Capture Screenshot")) {
				std::string name = screenshot_name;
				if (!name.length())
					name = "screenshot" + generate_unique_id();
				
				screenshot_context_t ctx{};
				ctx.name = name;
				ctx.prev_view_mode = view_mode;
				ctx.prev_white_light = white_light;
				ctx.prev_render_props = render_props;

				if (force_white_light)
					white_light = true;
				if (capture_pattern) {
					ctx.update_camera = true;
					ctx.prev_camera_pos = camera.position;
					ctx.prev_camera_rot = camera.rotation;
				}
				view_mode |= 1 << 8;
				render_props = false;

				if (ctx.update_camera)
					viewChanged();
				should_rebuild_pipelines = true;
				screenshot_context = ctx;
			}
		}

		if (should_rebuild_pipelines)
		{
			updateUniformBuffers();

			// Need to rebuild pipeline object with new specialization constant. Wait for rendering to finish.
			vkDeviceWaitIdle(device);
			preparePipelines();
			reBuildCommandBuffers();
		}
	}

	float inverse_smoothstep(float x) { return .5f - glm::sin(glm::asin(1.f-2.f*x)/3.f); }
};

VULKAN_EXAMPLE_MAIN()

