/*
*
*	Example implementation for "Accurate Rendering of Liquid-Crystals and Inhomogeneous Optically Anisotropic Media"
*	 by Shlomi Steinberg
*	 http://ssteinberg.xyz
*
*
*    This program is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*
*/

#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <filesystem>
#include <future>
#include <algorithm>
#include <atomic>
#include <regex>


// MATLAB engine
#ifdef _Matlab_Engine
#include <MatlabEngine.hpp>
#include <MatlabDataArray.hpp>
#else
#pragma message ("Warning: Matlab Engine not available.")
#endif

static constexpr int current_version = 0;
static const std::filesystem::path spirv_src[current_version+1] = { "../shaders/render/sem_v0.frag" };
static const std::filesystem::path spirv_dst = "../tmp/shader.spv";
static const std::filesystem::path profiles_directory = "../profiles";

inline std::string generate_unique_id() {
	auto t = std::time(nullptr);
	auto tm = *std::localtime(&t);

	std::ostringstream oss;
	oss << std::put_time(&tm, "%d-%m-%Y %H-%M-%S");
	return oss.str();
}

class profiles_manager {
	/***
	 * 	Profiles and profile generation job system
	 */

public:
	using version_t = int;
	enum class profile_type_t : int {
		slab = 0,
		droplet = 1,
		microdroplet = 2,
		droplet_axial = 3,
		droplet_pi_wall = 4,
		shell = 6,
		heat = 7,
	};
	static std::string profile_type_name(profile_type_t type) {
		switch (type) {
		default:
		case profile_type_t::slab:				return "slab";
		case profile_type_t::droplet:			return "droplet";
		case profile_type_t::microdroplet:		return "microdroplet";
		case profile_type_t::droplet_axial:		return "microdroplet-axial";
		case profile_type_t::droplet_pi_wall:	return "droplet-pi-wall";
		case profile_type_t::shell:				return "shell";
		case profile_type_t::heat:				return "heat-conduction";
		}
	}
	static std::pair<profile_type_t, bool> profile_type_for_name(const std::string &s) {
		if (s == "slab")				return { profile_type_t::slab, true };
		if (s == "droplet")				return { profile_type_t::droplet, true };
		if (s == "microdroplet")		return { profile_type_t::microdroplet, true };
		if (s == "microdroplet-axial")	return { profile_type_t::droplet_axial, true };
		if (s == "droplet-pi-wall")		return { profile_type_t::droplet_pi_wall, true };
		if (s == "shell")				return { profile_type_t::shell, true };
		if (s == "heat-conduction")		return { profile_type_t::heat, true };
		return { profile_type_t::slab, false };
	}
	
	struct profile_t {
		std::string id, name;
		std::vector<profile_type_t> types;
		version_t version;
		// Generated c methods
		std::string Ig1,Ig2,Ig_ce1,Ig_ce2,h1r,h1i,h2r,h2i,a,b,c,eo,ee,sigma_o,sigma_e;

		operator bool() const noexcept { return !!name.length(); }
	};
	using stages_t = std::deque<std::pair<std::string, std::string>>;
	struct profile_job_t {
		profile_t profile{};
		stages_t stages;	// Pairs of stage name and script

		bool should_optimize{ true };
		bool async{ true };
		std::chrono::high_resolution_clock::time_point optimization_start;

#ifdef _Matlab_Engine
		std::unique_ptr<matlab::engine::MATLABEngine> matlab;
		matlab::engine::FutureResult<std::unique_ptr<matlab::engine::MATLABEngine>> matlabFuture;
		std::optional<matlab::engine::FutureResult<void>> matlabEvalFuture;
		std::shared_ptr<std::basic_stringbuf<char16_t>> matlabOutput;
		std::shared_ptr<std::basic_stringbuf<char16_t>> matlabError;
#endif
		std::optional<std::future<bool>> routine_writeout;

		std::string status_text;
		enum class status_t {
			pending, successful, error
		} status = status_t::pending;
	};

	std::vector<profile_t> profiles;
	std::vector<profile_job_t> profile_jobs;

private:
	static constexpr auto gen_output_directory = "gen";
	static constexpr auto opt_output_directory = "opt";

	static version_t load_profile_version(const std::filesystem::path &p) {
		if (!std::filesystem::exists(p))
			return 0;
		return static_cast<version_t>(std::strtol(load_profile_routine(p).c_str(), nullptr, 10));
	}
	static std::pair<bool, version_t> check_profile_import(const std::filesystem::path &p) {
		// Checks that all required files exist
		const auto version = load_profile_version(p/"version");
		if (version < 0 || version > current_version) {
			std::cerr << "Unsupported version " << std::to_string(version) << " for profile " << p.string() << std::endl; 
			return { false, {} };
		}
		if (!std::filesystem::exists(p/"Ig1.c")  ||
			!std::filesystem::exists(p/"Ig2.c")  ||
			!std::filesystem::exists(p/"Ig_ce1.c") ||
			!std::filesystem::exists(p/"Ig_ce2.c") ||
			!std::filesystem::exists(p/"h1r.c")  ||
			!std::filesystem::exists(p/"h1i.c")  ||
			!std::filesystem::exists(p/"h2r.c")  ||
			!std::filesystem::exists(p/"h2i.c")  ||
			!std::filesystem::exists(p/"a.c")   ||
			!std::filesystem::exists(p/"b.c")   ||
			!std::filesystem::exists(p/"c.c")   ||
			!std::filesystem::exists(p/"eo.c")  ||
			!std::filesystem::exists(p/"ee.c")  ||
			!std::filesystem::exists(p/"sigma_o.c") ||
			!std::filesystem::exists(p/"sigma_e.c") ||
			!std::filesystem::exists(p/"type"))
			return { false, {} };

		return { true, version };
	}
	static bool copy_profile(const std::filesystem::path &src, const std::filesystem::path &dst) {
		// Checks that all required files exist
		return std::filesystem::copy_file(src/"Ig1.c",      dst/"Ig1.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"Ig2.c",      dst/"Ig2.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"Ig_ce1.c",   dst/"Ig_ce1.c",  std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"Ig_ce2.c",   dst/"Ig_ce2.c",  std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"h1r.c",      dst/"h1r.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"h1i.c",      dst/"h1i.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"h2r.c",      dst/"h2r.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"h2i.c",      dst/"h2i.c",     std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"a.c",        dst/"a.c",       std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"b.c",        dst/"b.c",       std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"c.c",        dst/"c.c",       std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"eo.c",       dst/"eo.c",      std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"ee.c",       dst/"ee.c",      std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"sigma_o.c",  dst/"sigma_o.c", std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"sigma_e.c",  dst/"sigma_e.c", std::filesystem::copy_options::overwrite_existing)  &&
			   std::filesystem::copy_file(src/"type",       dst/"type",      std::filesystem::copy_options::overwrite_existing);
	}
	static void create_profile_folder_structure(const std::filesystem::path &path) {
		std::filesystem::create_directory(path);
		std::filesystem::create_directory(path/gen_output_directory);
	}
	static std::pair<bool, version_t> check_profile(const std::filesystem::path &p) {
		// Checks that all required files exist
		if (!std::filesystem::exists(p/"name"))
			return { false, {} };
		return check_profile_import(p);
	}
	static std::string get_profile_name(const std::filesystem::path &p) {
		std::ifstream f(p/"name");
		if (!f)
			return "";
		std::string ret;
		std::getline(f, ret);
		return ret;
	}
	static std::vector<profile_type_t> get_profile_types(const std::filesystem::path &p) {
		std::ifstream f(p/"type");
		if (!f) {
			std::cerr << "Profile type unavilable for '" << p.string() << "'" << std::endl;
			return { profile_type_t::slab };
		}

		std::vector<profile_type_t> types;
		std::string t;
		while (std::getline(f, t)) {
			if (!t.length())
				break;
			const auto &r = profile_type_for_name(t);
			if (r.second)
				types.emplace_back(r.first);
		}
		if (types.size())
			return types;

		std::cerr << "Profile '" << p.string() << "' has unknown type. Defaulting to 'slab'." << std::endl;
		return { profile_type_t::slab };
	}
	static std::string load_profile_routine(const std::filesystem::path &p) {
		std::ifstream f(p);
		std::string str;
		// Reserve string size
		f.seekg(0, std::ios::end);   
		str.reserve(f.tellg());
		f.seekg(0, std::ios::beg);
		// Read file
		str.assign((std::istreambuf_iterator<char>(f)),
					std::istreambuf_iterator<char>());
		return str;
	}
	stages_t generate_matlab_script(std::filesystem::path output,
									std::uint32_t order,  
									const std::string& a, const std::string& b, const std::string& c, const std::string& eo, const std::string& ee, const std::string& so, const std::string& se) {
		// Load MATLAB script
		std::ifstream ifms("../scripts/gen.m");
		if (!ifms) {
			std::cerr << "MATLAB script couldn't be opened." << std::endl;
			return {};
		}
		stages_t matlab_scripts = { std::make_pair(std::string{"Preparing..."},std::string{}) };
		std::string line;
		bool input_injected = false;
		while (std::getline(ifms, line)) {
			const auto stage_str = std::string{ "%%% STAGE: " };
			if (line.compare(0, stage_str.length(), stage_str) == 0) {
				const auto stage = line.substr(stage_str.length()) + "...";
				matlab_scripts.emplace_back(std::make_pair(std::move(stage),std::string{}));
				continue;
			}
			if (!input_injected && line == "%$$ INPUT") {
				// Inject input
				matlab_scripts.back().second += "Order = " + std::to_string(order) + ";\n" + 
												"ProfileVersion = " + std::to_string(current_version) + ";\n" + 
												"StoreOutputAt = \"" + std::filesystem::absolute(output).string() + "/\";\n" +
												"a(h) = " + a + ";\n" + 
												"b(h) = " + b + ";\n" + 
												"c(h) = " + c + ";\n" + 
												"eo = " + eo + ";\n" + 
												"ee(h) = " + ee + ";\n" 
												"sigma_o = " + so + ";\n" + 
												"sigma_e(h) = " + se + ";\n";
				while (std::getline(ifms, line) && line!="%$$ /INPUT") {}
				input_injected = line=="%$$ /INPUT";
			}
			else
				matlab_scripts.back().second += line+"\n";
		}

		return input_injected ? matlab_scripts : stages_t{};
	}

public:
	void reload_profiles() {
		std::vector<std::future<profile_t>> profile_reads;
		for(auto &p : std::filesystem::directory_iterator(profiles_directory)) {
			// Check integrity
			const auto check_ret = check_profile(p.path());
			if (!check_ret.first) continue;

			profile_reads.emplace_back(std::async([p=p.path(), ver=check_ret.second]() {
				profile_t profile;
				profile.id   	= p.filename().string();
				profile.name 	= get_profile_name(p);
				profile.types   = get_profile_types(p);
				profile.version = ver;
				profile.Ig1   	= load_profile_routine(p/"Ig1.c");
				profile.Ig2   	= load_profile_routine(p/"Ig2.c");
				profile.Ig_ce1 	= load_profile_routine(p/"Ig_ce1.c");
				profile.Ig_ce2 	= load_profile_routine(p/"Ig_ce2.c");
				profile.h1r  	= load_profile_routine(p/"h1r.c");
				profile.h1i  	= load_profile_routine(p/"h1i.c");
				profile.h2r  	= load_profile_routine(p/"h2r.c");
				profile.h2i  	= load_profile_routine(p/"h2i.c");
				profile.a    	= load_profile_routine(p/"a.c");
				profile.b    	= load_profile_routine(p/"b.c");
				profile.c    	= load_profile_routine(p/"c.c");
				profile.eo   	= load_profile_routine(p/"eo.c");
				profile.ee   	= load_profile_routine(p/"ee.c");
				profile.sigma_o = load_profile_routine(p/"sigma_o.c");
				profile.sigma_e = load_profile_routine(p/"sigma_e.c");

				// Count ops
				const auto ops = 
					count_routine_ops(profile.Ig1) + 
					count_routine_ops(profile.Ig2) +   
					count_routine_ops(profile.Ig_ce1) +
					count_routine_ops(profile.Ig_ce2) +
					count_routine_ops(profile.h1r) +  	
					count_routine_ops(profile.h1i) +  	
					count_routine_ops(profile.h2r) +  	
					count_routine_ops(profile.h2i) +  	
					count_routine_ops(profile.a) +    	
					count_routine_ops(profile.b) +    	
					count_routine_ops(profile.c) +    	
					count_routine_ops(profile.eo) +   	
					count_routine_ops(profile.ee) +   	
					count_routine_ops(profile.sigma_o) +
					count_routine_ops(profile.sigma_e);
				std::ostringstream opsstrstrm;
				ops < 1000 ?
					opsstrstrm << ops :
					opsstrstrm << std::fixed << std::setprecision(1) << double(ops)/1e+3;
				const auto opsstr = opsstrstrm.str() + (ops<1000 ? " dops" : " kdops");
				// And append to name
				profile.name += " (" + opsstr + ")";
				
				return profile;
			}));
		}

		profiles = {};
		profiles.reserve(profile_reads.size());
		for (auto &f : profile_reads) {
			auto profile = f.get();
			if (!profile) continue;
			profiles.emplace_back(std::move(profile));
		}
		// Sort profiles by name
		std::sort(profiles.begin(),profiles.end(), [](const auto &lhs, const auto &rhs) { return lhs.name<rhs.name; });
	}
	static void delete_profile(const profile_t &p) {
		const auto path = profiles_directory / p.id;
		std::cout << "Deleting '" << path.string() << "'..." << std::endl;
		std::filesystem::remove_all(path);
	}
	
	static std::size_t count_routine_ops(const std::string &s) {
		// Optimized profiles are a GIMPLE dump therefore consist of a single operation on each line, minus declarations, close enough.
		const auto line_count = std::count_if(s.begin(), s.end(), [](auto c) { return c == '\n'; });
		return std::max<std::size_t>(3, line_count) - 3;
	}
	
	bool start_new_profile_job(std::uint32_t order, bool should_optimize, std::string name,
							   std::string a, std::string b, std::string c, std::string eo, std::string ee, std::string so, std::string se) {
#ifdef _Matlab_Engine
		profile_job_t job;
		job.profile.name = std::move(name);

		// Start MATLAB engine
		job.profile.id = generate_unique_id();
		while (std::filesystem::exists(profiles_directory/job.profile.id))
			job.profile.id += ".0";
		create_profile_folder_structure(profiles_directory/job.profile.id);
		job.stages = generate_matlab_script(profiles_directory/job.profile.id/gen_output_directory, order, a, b, c, eo, ee, so, se);
		if (!job.stages.size())
			return false;
		job.matlabFuture = matlab::engine::startMATLABAsync();
		job.status_text = "Starting MATLAB engine...";

		job.should_optimize = should_optimize;
		
		std::string debug_output{};
		for (const auto &s : job.stages)
			debug_output += "%% STAGE: " + s.first + "\n" + s.second;
		std::ofstream(profiles_directory/job.profile.id/"script.m") << debug_output;

		profile_jobs.emplace_back(std::move(job));
#endif

		return true;
	}
	bool start_new_import_job(std::string name, const std::filesystem::path &path, bool async = true) {
		profile_job_t job;
		job.profile.name = std::move(name);
		job.profile.id = generate_unique_id();
		job.async = async;

		const auto check_ret = check_profile_import(path);
		bool ok = false;
		if (!check_ret.first) {
			std::cerr << "Profile not found at provided path!" << std::endl;
			job.status_text = "Profile not found at path!";
			job.status = profile_job_t::status_t::error;
		}
		else if (check_ret.second != current_version) {
			job.status_text = "Profile version mismatch!";
			job.status = profile_job_t::status_t::error;
		}
		else {
			job.status_text = "Importing...";
			job.status = profile_job_t::status_t::pending;
			while (std::filesystem::exists(profiles_directory/job.profile.id))
				job.profile.id += ".0";
			create_profile_folder_structure(profiles_directory/job.profile.id);
			copy_profile(path, profiles_directory/job.profile.id/gen_output_directory);
			job.profile.types = get_profile_types(path);

			finish_job(job);
			ok = true;
		}

		profile_jobs.emplace_back(std::move(job));

		return ok;
	}
	bool optimize_routine(const std::filesystem::path &src, const std::filesystem::path &dst) {
#ifdef _linux
		auto opt_cmd = "g++ -w -O3 -fno-math-errno -ffinite-math-only -fno-rounding-math -fno-signaling-nans -fcx-limited-range -fno-trapping-math -I../gimple_dump_gcc_plugin/ -fplugin=../gimple_dump_gcc_plugin/gimple_dump.so \"" + src.string() + "\" -c -o /dev/null 1> \"" + dst.string() + "\"";
		return system(opt_cmd.c_str()) == 0;
#else
		return false;
#endif
	}
	bool prepare_routine(const std::filesystem::path &src, 
						 const std::filesystem::path &opt, 
						 const std::filesystem::path &dst, 
						 const std::string &name, 
						 bool complex_ret = false,
						 bool Ig = false,
						 bool optimize = true) {
		const auto filename = name+".c";
		auto p = src/filename;

		if (optimize) {
			std::filesystem::create_directory(opt);
			const auto opt_dst = opt/filename;

			// Run through gcc's optimizer
			if (!optimize_routine(p, opt_dst)) {
				std::cerr << "Optimization failed" << std::endl;
				return false;
			}

			p = opt_dst;
		}

		// read output and generate GLSL
		std::ifstream f(p);
		if (!f)
			return false;

		std::string line, content;

		std::string ret = complex_ret ? "dvec2" : "double";

		std::size_t lines = 0;
		while (std::getline(f, line)) {
			if (line.length() && line[0]=='#') continue;
			if (line.length()>=4 && line.substr(0,4)=="auto") continue;

			// Replace 'pow'/'exp' with our custom 'powd'/'expd' that emulates double precision arithmetics
			line = std::regex_replace(line, std::regex("<double>"), "");
			line = std::regex_replace(line, std::regex("std::"), "");
			line = std::regex_replace(line, std::regex(".constprop"), "");
			line = std::regex_replace(line, std::regex("pow"), "powd");
			line = std::regex_replace(line, std::regex("exp"), "expd");
			line = std::regex_replace(line, std::regex("sqrt"), "sqrtd");
			line = std::regex_replace(line, std::regex("complex"), "dvec2");
			// Replace gcc generated __builtin_* functions
			if (optimize)
				line = std::regex_replace(line, std::regex("__builtin_"), "");
			// Add the 'lf' double-precision decorator to floating-point literals
			const std::string lf = "lf";
			std::match_results<std::string::iterator> m;
			auto it = line.begin();
			while (std::regex_search(it, line.end(), m, std::regex(R"(\W[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")))
				it=line.insert(m[0].second, lf.begin(), lf.end())+2;

			content += (line != "}" ? "\t" : "") + line + "\n";
			++lines;
		}
		// When Ig1/Ig2 return a constant (usually 0) change the return type.
		if (lines<5)
			ret = "double";

		// Write out
		{
			std::ofstream of{ dst/filename };
			of << ret + " " + name + "(const double K" + (Ig ? ", const dvec2 Ga, const dvec2 Gb, const dvec2 erf0" : "") + ", const double h, const double c0, const double c1, const double c2, const double c3, const double c4) {" << std::endl;
			of << content;
		}

		std::cout << "Routine " << name << " parsed successfully.";
		std::cout << std::endl;

		return true;
	}
	void finish_job(profile_job_t &p, const char* exception = nullptr) {
		const auto& src = profiles_directory / p.profile.id / gen_output_directory;
		const auto& opt = profiles_directory / p.profile.id / opt_output_directory;
		const auto& dst = profiles_directory / p.profile.id;
		const auto& err =
#ifdef _Matlab_Engine
			p.matlabError  ? matlab::engine::convertUTF16StringToUTF8String(p.matlabError->str())  :
#endif
			std::string{};
		const auto& log =
#ifdef _Matlab_Engine
			p.matlabOutput ? matlab::engine::convertUTF16StringToUTF8String(p.matlabOutput->str()) :
#endif
			std::string{};

		// Write out MATLAB log
		std::ofstream(src/"matlab.log") << log;

#ifdef _Matlab_Engine
		p.matlab = nullptr;
		p.matlabFuture = {};
		p.matlabEvalFuture = {};
#endif
		// Check for errors
		if (err.length() || exception) {
			p.status = profile_job_t::status_t::error;
			std::cout << "Generation of profile '" << p.profile.name << "' finished with errors." << std::endl;
			std::cerr << err << std::endl;
			if (exception)
				std::cerr << exception << std::endl;
			return;
		}

		const bool should_optimize = p.should_optimize;
		p.status_text = "Generating routines...";
		p.optimization_start = std::chrono::high_resolution_clock::now();
		p.routine_writeout = std::async(std::launch::async, [=, async_policy = p.async, name = p.profile.name]() -> bool {
			// Generate routines
			bool success = true;
			const auto policy = async_policy ? std::launch::async : std::launch::deferred;
			for (auto& f : std::array{
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "Ig1",		true,  true,  should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "Ig2",		true,  true,  should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "Ig_ce1",	false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "Ig_ce2",	false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "h1r",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "h1i",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "h2r",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "h2i",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "a", 		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "b", 		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "c", 		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "eo",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "ee",		false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "sigma_o", false, false, should_optimize); }),
				std::async(policy, [&]() { return prepare_routine(src, opt, dst, "sigma_e", false, false, should_optimize); })}) {
				success &= f.get();
			}
			if (!success) {
				std::cout << "Generation of profile '" << name << "' finished with errors." << std::endl;
				std::cerr << "Couldn't prepare routines at '" << dst.string() << "'." << std::endl;
				return false;
			}
			return success;
		});
	}
	void stop_job(profile_job_t &p) {
		if (p.status == profile_job_t::status_t::pending) {
#ifdef _Matlab_Engine
			// Interrupt command
			if (!p.matlab)
				p.matlabFuture.cancel();
			else if (p.matlabEvalFuture)
				p.matlabEvalFuture.value().cancel();
#endif
			std::cout << "Generation of profile '" << p.profile.name << "' interrupted." << std::endl;
			p.status = profile_job_t::status_t::error;
			p.status_text = "Aborted";
		}

#ifdef _Matlab_Engine
		p.matlab = nullptr;
		p.matlabFuture = {};
		p.matlabEvalFuture = {};
#endif
	}
	bool update_jobs() {
		bool new_profile = false;
		for (auto &p : profile_jobs) {
			if (p.status != profile_job_t::status_t::pending) continue;

			if (p.routine_writeout) {
				if (p.routine_writeout.value().wait_for(std::chrono::seconds(0)) == std::future_status::ready) {
					if (p.routine_writeout.value().get()) {

						if (p.should_optimize) {
							// Calculate optimization time
							const auto duration = std::chrono::high_resolution_clock::now() - p.optimization_start;
							std::cout << "Optimization time: " << std::chrono::duration_cast<std::chrono::milliseconds>(duration).count()/1000. << "." << std::endl;
						}

						std::string types;
						for (const auto &type : p.profile.types)
							types += profile_type_name(type) + "\n";

						// Write profile name, types and version
						std::ofstream(profiles_directory/p.profile.id/"name") << p.profile.name;
						std::ofstream(profiles_directory/p.profile.id/"version") << std::to_string(current_version);
						std::ofstream(profiles_directory/p.profile.id/"type") << types;

						p.status_text = "Done";
						p.status = profile_job_t::status_t::successful;
						std::cout << "Generation of profile '" << p.profile.name << "' finished successfully." << std::endl;

						new_profile = true;
					}
					else {
						p.status_text = "Failed";
						p.status = profile_job_t::status_t::error;
					}
					p.routine_writeout = {};
				}
			}
#ifdef _Matlab_Engine
			else if (!p.matlab) {
				// Check if MATLAB engine has started
				if (p.matlabFuture.wait_for(std::chrono::seconds(0)) == std::future_status::ready) {
					try {
						p.matlab = p.matlabFuture.get();
					} catch(std::exception e) {
						std::cerr << e.what() << std::endl;
						p.status_text = e.what();
						finish_job(p, e.what());
						continue;
					}
				}
			}
			if (!!p.matlab && !p.matlabEvalFuture) {
				// Start evaluation
				// Change root dir first to MATLAB scripts dir
				p.matlabOutput = std::make_shared<std::basic_stringbuf<char16_t>>();
				p.matlabError  = std::make_shared<std::basic_stringbuf<char16_t>>();
				p.matlabEvalFuture = p.matlab->evalAsync(matlab::engine::convertUTF8StringToUTF16String(std::string("cd ") + "../scripts/matlab/"),
														 p.matlabOutput, p.matlabError);
				p.status_text = "Starting evaluation of MATLAB script...";
			}
			else if (p.matlabEvalFuture) {
				// Check if evaluation is ready
				if (p.matlabEvalFuture.value().wait_for(std::chrono::seconds(0)) == std::future_status::ready) {
					try {
						p.matlabEvalFuture.value().get();
					} catch(std::exception e) {
						std::cerr << e.what() << std::endl;
						p.status_text = e.what();
						finish_job(p, e.what());
						continue;
					}

					// Do we have more stages?
					if (p.stages.size()) {
						p.matlabEvalFuture = p.matlab->evalAsync(matlab::engine::convertUTF8StringToUTF16String(p.stages.front().second),
																 p.matlabOutput, p.matlabError);
						p.status_text = std::move(p.stages.front().first);
						p.stages.pop_front();
					}
					else
						finish_job(p);
				}
			}
#endif
		}

		return new_profile;
	}
};
