function [ro,re,to,te] = fresnel_aniso(K,qi,Ei,eo,eo2,ee,ee2,A,A2)
% function [ro,re,to,te] = fresnel_aniso(K,qi,Ei,eo,eo2,ee,ee2,A,A2)
% Computes the fresnel coefficients for an anisotropic-to-anisotropic
% interface
% K is the incidence parameter
% qi is the incident wave's normal component
% Ei is the incident wave's electric field
% eo is ordinary index-of-refraction of the source medium
% eo2 is ordinary index-of-refraction of the destination medium
% ee is extraordinary index-of-refraction of the source medium
% ee2 is extraordinary index-of-refraction of the destination medium
% A is the optic axis of the source medium
% A2 is the optic axis of the destination medium


    [Eop, ~] = E_o(eo,A,K);
    [Eep, ~] = E_e(eo,ee,A,K);
    qop = qo(eo,K);
    [qep, ~] = qe(eo,ee,A,K);
    [~, E2om] = E_o(eo2,A2,K);
    [~, E2em] = E_e(eo2,ee2,A2,K);
    q2om = -qo(eo2,K);
    [~, q2em] = qe(eo2,ee2,A2,K);
    

    Ei_x = Ei(1);
    Ei_y = Ei(2);
    Ei_z = Ei(3);
    Eop_x = Eop(1);
    Eop_y = Eop(2);
    Eop_z = Eop(3);
    Eep_x = Eep(1);
    Eep_y = Eep(2);
    Eep_z = Eep(3);
    E2om_x = E2om(1);
    E2om_y = E2om(2);
    E2om_z = E2om(3);
    E2em_x = E2em(1);
    E2em_y = E2em(2);
    E2em_z = E2em(3);
    
    C_1 = q2em - q2om;
    C_2 = qep - qi;
    C_3 = q2em - qi;
    C_4 = q2em - qep;
    C_5 = q2om - qi;
    C_6 = q2om - qep;
    C_7 = qep - qop;
    C_8 = q2em - qop;
    C_9 = q2om - qop;

    Nro= ...
    (E2em_x*E2om_x*C_1*(Eep_y*Eop_z*K - Eep_z*Eop_y*K + Eep_z*Eop_z*C_7) ...
    + E2em_x*E2om_y*K*(Eep_z*Eop_x*C_8 - Eep_x*Eop_z*C_4) ...
    + E2em_x*E2om_z*(Eop_x*C_8*(Eep_z*C_6 - Eep_y*K) + Eep_x*C_4*(Eop_y*K - Eop_z*C_9)) ...
    + E2em_y*E2om_x*K*(Eep_x*Eop_z*C_6 - Eep_z*Eop_x*C_9)  ...
    + Eep_x*Eop_x*C_7*(K*E2em_y*E2om_z - K*E2em_z*E2om_y + E2em_z*E2om_z*C_1) ...
    + E2em_z*E2om_x*(Eep_x*C_6*(Eop_z*C_8 - Eop_y*K) + Eop_x*C_9*(Eep_y*K - Eep_z*C_4)));

    ro= ...
    (-E2em_x*E2om_y*K*(Eep_z*Ei_x*C_3 - Eep_x*Ei_z*C_4) ...
    + E2em_x*E2om_z*(Eep_x*C_4*(Ei_z*C_5 - Ei_y*K) - Ei_x*C_3*(Eep_z*C_6 - Eep_y*K)) ...
    - E2em_y*E2om_x*K*(Eep_x*Ei_z*C_6 - Eep_z*Ei_x*C_5) ...
    - Eep_x*Ei_x*K*C_2*(E2em_y*E2om_z - E2em_z*E2om_y) ...
    + E2em_z*E2om_x*(Ei_x*C_5*(Eep_z*C_4 - Eep_y*K) - Eep_x*C_6*(Ei_z*C_3 - Ei_y*K)) ...
    - E2em_z*E2om_z*(Eep_x*Ei_x*q2em*C_2 - Eep_x*Ei_x*q2om*C_2) ...
    + E2em_x*E2om_x*C_1*(Eep_z*Ei_y*K - Eep_y*Ei_z*K - Eep_z*Ei_z*C_2)) ...
    / ...
    Nro;

    re = ...
    -(E2em_x*E2om_z*q2em*(Ei_z*C_5 - Ei_y*K - (Eop_y*K - Eop_z*C_9)*ro)  ...
    - E2em_z*E2om_x*q2om*(Ei_z*C_3 - Ei_y*K - (Eop_y*K - Eop_z*C_8)*ro)  ...
    + E2em_z*E2om_z*C_1*(Ei_x*qi + Eop_x*qop*ro)  ...
    - K*(Ei_z + Eop_z*ro)*(E2em_y*E2om_x*q2om - E2em_x*E2om_y*q2em) ...
    - K*(Ei_x*qi + Eop_x*qop*ro)*(E2em_z*E2om_y - E2em_y*E2om_z)) ...
    / ...
    (E2om_z*(E2em_x*q2em*(Eep_z*C_6 - Eep_y*K) + Eep_x*qep*(E2em_z*C_1 + E2em_y*K)) ...
    - E2om_x*q2om*(E2em_z*(Eep_z*C_4 - Eep_y*K) + E2em_y*Eep_z*K) ...
    - E2om_y*K*(E2em_z*Eep_x*qep - E2em_x*Eep_z*q2em));


    Nt = E2em_x*E2om_z - E2em_z*E2om_x;
    D_1 = Ei_z + Eep_z*re + Eop_z*ro;
    D_2 = Ei_x + Eep_x*re + Eop_x*ro;

    to = (E2em_x*D_1 - E2em_z*D_2)/Nt;
    te = (E2om_z*D_2 - E2om_x*D_1)/Nt;
end

